from deprecated.sphinx import versionadded


@versionadded(
    version="2.3.1",
    reason="Added to avoid display of unrelevant warning to the end user",
)
class SepalWarning(Warning):
    """
    A custom warning class that will be the only one to be displayed in the Alert in voila.
    The other normal warning such as lib DeprecationWarning will be displayed in the notebook but hidden to the end user
    """

    pass



from pathlib import Path
from traitlets import link

import ipyvuetify as v

from sepal_ui import sepalwidgets as sw
from sepal_ui.message import ms
from sepal_ui import reclassify as rec

__all__ = ["ReclassifyTile"]


class ReclassifyTile(sw.Tile):
    """

    All in one tile to reclassify GEE assets or local raster and create custom classifications

    Args:
        results_dir (str|pathlike object): Directory to store the outputs (rasters, and csv_files). default to ~/downloads
        gee (bool): Use GEE variant, to reclassify assets or local input. default True
        dst_class (str|pathlib.Path, optional): the file to be used as destination classification. for app that require specific code system the file can be set prior and the user won't have the oportunity to change it
        default_class (dict|optional): the default classification system to use, need to point to existing sytem: {name: absolute_path}
        folder(str, optional): the init GEE asset folder where the asset selector should start looking (debugging purpose)
    """

    result_dir = None
    "pathlib.Path: Directory to store the outputs (rasters, and csv_files)."

    model = None
    "ReclassifyModel: the reclassify model to use with these inputs"

    reclassify_view = None
    "ReclassifyView: a fully qualified ReclassifyView object"

    table_view = None
    "TableView: a fully qualified TableView object"

    def __init__(
        self,
        results_dir=Path.home() / "downloads",
        gee=True,
        dst_class=None,
        default_class={},
        aoi_model=None,
        folder=None,
        **kwargs
    ):

        # output directory
        self.results_dir = Path(results_dir)

        self.aoi_model = aoi_model
        # create the model
        self.model = rec.ReclassifyModel(
            dst_dir=self.results_dir, gee=gee, aoi_model=self.aoi_model, folder=folder
        )

        # set the tabs elements
        self.reclassify_view = rec.ReclassifyView(
            self.model,
            out_path=self.results_dir,
            gee=gee,
            default_class=default_class,
            aoi_model=aoi_model,
            save=True,
            folder=folder,
        ).nest_tile()

        self.table_view = rec.TableView(out_path=self.results_dir).nest_tile()

        # create the tab
        tiles = [self.reclassify_view, self.table_view]

        self._tabs = v.Tabs(
            class_="mt-5",
            grow=True,
            # background_color='grey darken-4',
            v_model=0,
            children=[
                v.Tab(children=[t.title.children[0].children[0]], key=i)
                for i, t in enumerate(tiles)
            ],
        )

        self._content = v.TabsItems(
            v_model=0,
            children=[v.TabItem(children=[t], key=i) for i, t in enumerate(tiles)],
        )

        super().__init__(
            id_="reclassify_tile",
            title=ms.rec.tile.title,
            inputs=[self._tabs, self._content],
            **kwargs
        )

        # js interaction
        link((self._tabs, "v_model"), (self._content, "v_model"))



from traitlets import Unicode
from IPython.display import display
import ipyvuetify as v

# change vuetify theming
v.theme.dark = True

# set the colors for the dark theme
v.theme.themes.dark.primary = "#B3842E"
v.theme.themes.dark.accent = "#a1458e"
v.theme.themes.dark.secondary = "#324a88"
v.theme.themes.dark.success = "#3F802A"
v.theme.themes.dark.info = "#79B1C9"
v.theme.themes.dark.warning = "#b8721d"
v.theme.themes.dark.error = "#A63228"

# fixed colors
sepal_main = "#24221F"
sepal_darker = "#1a1a1a"

# set the background
bg_color = "#121212" if v.theme.dark else "#fff"


class Styles(v.VuetifyTemplate):
    """
    Fixed styles to avoid leaflet maps overlap sepal widgets
    """

    template = Unicode(
        """
        <style>
            .leaflet-pane {z-index : 2 !important;}
            .leaflet-top, .leaflet-bottom {z-index : 2 !important;}
            .leaflet-widgetcontrol {box-shadow: none}
            main.v-content {padding-top: 0px !important;}
        </style>
        <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.4/css/all.min.css"/>
    """
    ).tag(sync=True)
    "Unicode: the trait embeding the maps style"


styles = Styles()
display(styles)

COMPONENTS = {
    "PROGRESS_BAR": {
        "color": "indigo",
    }
}

# default styling of the aoi layer
AOI_STYLE = {
    "stroke": True,
    "color": v.theme.themes.dark.success,
    "weight": 2,
    "opacity": 1,
    "fill": True,
    "fillColor": v.theme.themes.dark.success,
    "fillOpacity": 0.4,
}


_folder = {"color": "amber", "icon": "far fa-folder"}
_table = {"color": "green accent-4", "icon": "far fa-table"}
_vector = {"color": "deep-purple", "icon": "far fa-vector-square"}
_other = {"color": "light-blue", "icon": "far fa-file"}
_parent = {"color": "white", "icon": "far fa-folder-open"}
_image = {"color": "deep-purple", "icon": "far fa-image"}

ICON_TYPES = {
    "": _folder,
    ".csv": _table,
    ".txt": _table,
    ".tif": _image,
    ".tiff": _image,
    ".vrt": _image,
    ".shp": _vector,
    ".geojson": _vector,
    "DEFAULT": _other,
    "PARENT": _parent,
}



import ipyvuetify as v
from markdown import markdown
from traitlets import Unicode, Any
from ipywidgets import link
from deprecated.sphinx import versionadded

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget

__all__ = ["Markdown", "Tooltip", "CopyToClip"]


class Markdown(v.Layout, SepalWidget):
    """
    Custom Layout based on the markdown text given

    Args:
        mkd_str (str): the text to display using the markdown convention. multi-line string are also interpreted
        kwargs (optional): Any parameter from a v.Layout. If set, 'children' will be overwritten
    """

    def __init__(self, mkd_str="", **kwargs):

        mkd = markdown(mkd_str, extensions=["fenced_code", "sane_lists"])

        # need to be nested in a div to be displayed
        mkd = "<div>\n" + mkd + "\n</div>"

        # make every link to point to target black (to avoid nested iframe in sepal)
        mkd = mkd.replace("<a", '<a target="_blank"')

        # create a Html widget
        class MyHTML(v.VuetifyTemplate):
            template = Unicode(mkd).tag(sync=True)

        content = MyHTML()

        # set default parameters
        kwargs["row"] = kwargs.pop("row", True)
        kwargs["class_"] = kwargs.pop("class_", "pa-5")
        kwargs["align_center"] = kwargs.pop("align_center", True)
        kwargs["children"] = [v.Flex(xs12=True, children=[content])]

        # call the constructor
        super().__init__(**kwargs)


class Tooltip(v.Tooltip):
    """
    Custom widget to display tooltip when mouse is over widget

    Args:
        widget (Vuetify.widget): widget used to display tooltip
        tooltip (str): the text to display in the tooltip
    """

    def __init__(self, widget, tooltip, *args, **kwargs):

        self.v_slots = [
            {"name": "activator", "variable": "tooltip", "children": widget}
        ]
        widget.v_on = "tooltip.on"

        self.children = [tooltip]

        super().__init__(*args, **kwargs)


@versionadded(version="2.2.0", reason="New clipping widget")
class CopyToClip(v.VuetifyTemplate):
    """
    Custom textField that provides a handy copy-to-clipboard javascript behaviour.
    When the clipboard btn is clicked the v_model will be copied in the local browser clipboard. You just have to change the clipboard v_model. when copied, the icon change from a copy to a check.

    Args:
        kwargs: any argument that can be used with a v.TextField
    """

    tf = None
    "v.TextField: the textfield widget that holds the v_model to copy"

    v_model = Any(None).tag(sync=True)
    "Any trait: a v_model trait that embed the string to copy"

    def __init__(self, **kwargs):

        # add the default params to kwargs
        kwargs["outlined"] = kwargs.pop("outlined", True)
        kwargs["label"] = kwargs.pop("label", "Copy To clipboard")
        kwargs["readonly"] = kwargs.pop("readonly", True)
        kwargs["append_icon"] = kwargs.pop("append_icon", "fas fa-clipboard")
        kwargs["v_model"] = kwargs.pop("v_model", None)
        kwargs["class_"] = kwargs.pop("class_", "ma-5")

        # set the default v_model
        self.v_model = kwargs["v_model"]

        # create component
        self.tf = v.TextField(**kwargs)
        self.components = {"mytf": self.tf}

        # template with js behaviour
        self.template = """
        <mytf/>
        <script>
            {methods: {
                jupyter_clip(_txt) {
                    var tempInput = document.createElement("input");
                    tempInput.value = _txt;
                    document.body.appendChild(tempInput);
                    tempInput.focus();
                    tempInput.select();
                    document.execCommand("copy");
                    document.body.removeChild(tempInput);
                }
            }}
        </script>"""

        super().__init__()

        # js behaviour
        self.tf.on_event("click:append", self._clip)
        link((self, "v_model"), (self.tf, "v_model"))

    def _clip(self, widget, event, data):
        self.send({"method": "clip", "args": [self.tf.v_model]})
        self.tf.append_icon = "fas fa-clipboard-check"

        return



from datetime import datetime

from ipywidgets import jslink
import ipyvuetify as v
from deprecated.sphinx import deprecated
from traitlets import Unicode, observe, directional_link, Bool

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget, TYPES

__all__ = ["Divider", "Alert", "StateBar"]


class Divider(v.Divider, SepalWidget):
    """
    A custom Divider with the ability to dynamically change color
    Whenever the type\_ trait is modified, the divider class will change accordingly

    Args:
        class\_ (str, optional): the initial color of the divider
        kwargs (optional): any parameter from a v.Divider. if set, 'class_' will be overwritten.
    """

    type_ = Unicode("").tag(sync=True)
    "str: Added type_ trait to specify the current color of the divider"

    def __init__(self, class_="", **kwargs):
        kwargs["class_"] = class_
        super().__init__(**kwargs)

    @observe("type_")
    def add_class_type(self, change):
        """
        Change the color of the divider according to the type\_
        It is binded to the type\_ traitlet but can also be called manually

        Args:
            change (dict): the only useful key is 'new' which is the new required color

        Return:
            self
        """

        type_ = change["new"]
        classes = self.class_.split(" ")
        existing = list(set(classes) & set(TYPES))
        if existing:
            classes[classes.index(existing[0])] = type_
            self.class_ = " ".join(classes)
        else:
            self.class_ += f" {type_}"

        return self


class Alert(v.Alert, SepalWidget):
    """
    A custom Alert widget. It is used as the output of all processes in the framework.
    In the voila interfaces, print statement will not be displayed.
    Instead use the sw.Alert method to provide information to the user.
    It's hidden by default.

    Args:
        type\_ (str, optional): The color of the Alert
        kwargs (optional): any parameter from a v.Alert. If set, 'type' will be overwritten.
    """

    def __init__(self, type_=None, **kwargs):

        # set default parameters
        kwargs["text"] = kwargs.pop("text", True)
        kwargs["type"] = type_ if (type_ in TYPES) else TYPES[0]
        kwargs["class_"] = kwargs.pop("class_", "mt-5")

        # call the constructor
        super().__init__(**kwargs)

        self.hide()

    def update_progress(self, progress, msg="Progress", bar_length=30):
        """
        Update the Alert message with a progress bar. This function will stay until we manage to use tqdm in the widgets

        Args:
            progress (float): the progress status in float [0, 1]
            msg (str, optionnal): The message to use before the progress bar
            bar_length (int, optionnal): the length of the progress bar in characters

        Return:
            self
        """

        # define the characters to use in the progress bar
        plain_char = "█"
        empty_char = " "

        # cast the progress to float
        progress = float(progress)
        if not (0 <= progress <= 1):
            raise ValueError(f"progress should be in [0, 1], {progress} given")

        # set the length parameter
        block = int(round(bar_length * progress))

        # construct the message content
        text = f"|{plain_char * block + empty_char * (bar_length - block)}|"

        # add the message to the output
        self.add_live_msg(
            v.Html(
                tag="span",
                children=[
                    v.Html(tag="span", children=[f"{msg}: "], class_="d-inline"),
                    v.Html(tag="pre", class_="info--text d-inline", children=[text]),
                    v.Html(
                        tag="span",
                        children=[f" {progress *100:.1f}%"],
                        class_="d-inline",
                    ),
                ],
            )
        )

        return self

    def add_msg(self, msg, type_="info"):
        """
        Add a message in the alert by replacing all the existing one.
        The color can also be changed dynamically

        Args:
            msg (str): the message to display
            type\_ (str, optional): the color to use in the widget

        Return:
            self
        """
        self.show()
        self.type = type_ if (type_ in TYPES) else TYPES[0]
        self.children = [v.Html(tag="p", children=[msg])]

        return self

    def add_live_msg(self, msg, type_="info"):
        """
        Add a message in the alert by replacing all the existing one.
        Also add the timestamp of the display.
        The color can also be changed dynamically

        Args:
            msg (str): the message to display
            type\_ (str, optional): the color to use in the widget

        Return:
            self
        """

        current_time = datetime.now().strftime("%Y/%m/%d, %H:%M:%S")

        self.show()
        self.type = type_ if (type_ in TYPES) else TYPES[0]

        self.children = [
            v.Html(tag="p", children=["[{}]".format(current_time)]),
            v.Html(tag="p", children=[msg]),
        ]

        return self

    def append_msg(self, msg, section=False, type_="info"):
        """
        Append a message in a new parragraph, with or without divider

        Args:
            msg (str): the message to display
            section (bool, optional): add a Divider before the added message

        Return:
            self
        """

        self.show()
        self.type = type_

        if len(self.children):
            current_children = self.children[:]
            if section:
                # As the list is mutable, and the trait is only triggered when
                # the children is changed, so have to create a copy.
                divider = Divider(class_="my-4", style="opacity: 0.22")

                # link Alert type with divider type
                directional_link((self, "type"), (divider, "type_"))
                current_children.extend([divider, v.Html(tag="p", children=[msg])])
                self.children = current_children

            else:
                current_children.append(v.Html(tag="p", children=[msg]))
                self.children = current_children
        else:
            self.add_msg(msg)

        return self

    def remove_last_msg(self):
        """
        Remove the last msg printed in the Alert widget

        Return:
            self
        """

        if len(self.children) > 1:
            current_children = self.children[:]
            self.children = current_children[:-1]
        else:
            self.reset()

        return self

    def reset(self):
        """
        Empty the messages and hide it

        Return:
            self
        """

        self.children = [""]
        self.hide()

        return self

    @deprecated(version="2.1.0", reason="use a Model object instead")
    def bind(self, widget, obj, attribute, msg=None, verbose=True, secret=False):
        """
        Bind the attribute to the widget and display it in the alert.
        The binded input need to have an active `v_model` trait.
        After the binding, whenever the `v_model` of the input is changed, the io attribute is changed accordingly.
        The value can also be displayed in the alert with a custom message with the following format = `[custom message] + [v_model]`

        Args:
            widget (v.XX): an ipyvuetify input element with an activated `v_model` trait
            obj (io): any io object
            attribute (str): the name of the attribute in io object
            msg (str, optionnal): the output message displayed before the variable
            verbose (bool, optional): wheter the variable should be displayed to the user
            secret (bool, optional): either if the variable is secret or not. If true only "*" will be shown in the output

        Return:
            self
        """
        if not msg:
            msg = "The selected variable is: "

        def _on_change(change, obj=obj, attribute=attribute, msg=msg):

            # if the key doesn't exist the getattr function will raise an AttributeError
            getattr(obj, attribute)

            # change the obj value
            setattr(obj, attribute, change["new"])

            # add the message if needed
            if secret:
                msg += "*" * len(str(change["new"]))
            else:
                msg += str(change["new"])

            if verbose:
                self.add_msg(msg)

            return

        widget.observe(_on_change, "v_model")

        return self

    def check_input(self, input_, msg=None):
        """
        Check if the inpupt value is initialized.
        If not return false and display an error message else return True

        Args:
            input\_ (any): the input to check
            msg (str, optionnal): the message to display if the input is not set

        Return:
            (bool): check if the value is initialized
        """
        if not msg:
            msg = "The value has not been initialized"
        init = True

        # check the collection type that are the only one supporting the len method
        try:
            init = False if len(input_) == 0 else init
        except Exception:
            init = False if input_ is None else init

        if init is False:
            self.add_msg(msg, "error")

        return init


class StateBar(v.SystemBar):

    """Widget to display quick messages on simple inline status bar

    Args:
       kwargs (optional): any parameter from a v.SystemBar. If set, 'children' will be overwritten.
    """

    msg = Unicode("").tag(sync=True)
    "str: the displayed message"

    loading = Bool(False).tag(sync=True)
    "bool: either or not the circular progress is spinning"

    progress = None
    "widget: The ProgressCircular widget that will be displayed in the statebar"

    def __init__(self, **kwargs):

        self.progress = v.ProgressCircular(
            indeterminate=self.loading,
            value=100,
            small=True,
            size=15,
            color="primary",
            class_="mr-2",
        )

        # set default parameter
        kwargs["children"] = [self.progress, self.msg]

        # call the constructor
        super().__init__(**kwargs)

        jslink((self, "loading"), (self.progress, "indeterminate"))

    @observe("loading")
    def _change_loading(self, change):
        """Change progress wheel state"""
        self.progress.indeterminate = self.loading

    @observe("msg")
    def _change_msg(self, change):
        """Change state bar message"""
        self.children = [self.progress, self.msg]

    def add_msg(self, msg, loading=False):
        """Change current status message"""
        self.msg = msg
        self.loading = loading

        return self



from pathlib import Path
from colorsys import rgb_to_hls, rgb_to_hsv
from traitlets import Int

import ipyvuetify as v
from matplotlib.colors import to_rgb
import pandas as pd

from sepal_ui.reclassify import parameters as param
from sepal_ui import sepalwidgets as sw
from sepal_ui.scripts import utils as su
from sepal_ui.message import ms

__all__ = ["TableView"]


class ClassTable(sw.DataTable):

    """
    Custom data table to modify, display and save classification. From this interface, a user can modify a classification starting from a scratch or by loading a classification file. the display datatable allow all the CRUD fonctionality (create, read, update, delete).

    Attributes:
        SCHEMA (const list): the schema of the classification: ['id', 'code', 'description', 'color']
        out_path (pathlib.Path): the output folder to save the created classifications
        save_dialog (v.Dialog): the Dialog to save the classification (select name and format)
        edit_dialog (v.Dialog): the dialog to edit/create a noew line in the data-table. The user will provide: the code, the description and the color (with a color selector)
        edit_icon (v.Icon): the btn to open the edit_dialog (a line need to be selected in the table)
        delete_icon (v.icon): the btn to delete a selected line in the table
        add_icon (v.Icon): the btn to add a new line in the table
        save_icon (v.Icon): the btn to open the save_dialog and save the current classification table

    Args:
        out_path (str|optional): output path where table will be saved, default to ~/downloads/
    """

    SCHEMA = param.SCHEMA

    def __init__(self, out_path=Path.home() / "downloads", **kwargs):

        # save the output path
        self.out_path = Path(out_path)

        # create the dialogs
        self.save_dialog = SaveDialog(
            table=self, out_path=self.out_path, transition=False
        )
        self.edit_dialog = EditDialog(table=self)

        # create the 4 CRUD btn
        # and set them in the top slot of the table
        self.edit_btn = sw.Btn(
            ms.rec.table.btn.edit,
            icon="fas fa-pencil-alt",
            class_="ml-2 mr-2",
            color="secondary",
            small=True,
        )
        self.delete_btn = sw.Btn(
            ms.rec.table.btn.delete, icon="fas fa-trash-alt", color="error", small=True
        )
        self.add_btn = sw.Btn(
            ms.rec.table.btn.add, icon="fas fa-plus", color="success", small=True
        )
        self.save_btn = sw.Btn(ms.rec.table.btn.save, icon="far fa-save", small=True)

        slot = v.Toolbar(
            class_="d-flex mb-6",
            flat=True,
            children=[
                self.edit_dialog,
                self.save_dialog,
                v.ToolbarTitle(children=["Actions"]),
                v.Divider(class_="mx-4", inset=True, vertical=True),
                v.Flex(
                    class_="ml-auto",
                    children=[self.add_btn, self.edit_btn, self.delete_btn],
                ),
                v.Divider(class_="mx-4", inset=True, vertical=True),
                self.save_btn,
            ],
        )

        self.v_slots = [{"name": "top", "variable": "top", "children": [slot]}]

        # set up default parameters of the datatable
        self.v_model = []
        self.items = []
        self.item_key = "id"
        self.show_select = True
        self.single_select = True
        self.hide_default_footer = True
        self.headers = [
            {"text": v[0].capitalize(), "value": t} for t, v in self.SCHEMA.items()
        ]

        # create the object
        super().__init__(**kwargs)

        # js events
        self.edit_btn.on_event("click", self._edit_event)
        self.delete_btn.on_event("click", self._remove_event)
        self.add_btn.on_event("click", self._add_event)
        self.save_btn.on_event("click", self._save_event)

    def populate_table(self, items_file=None):
        """
        Populate table. It will fill the table with the item contained in the items_file parameter. If no file is provided the table is reset.

        Args:
            items (Path object|optional): file containing classes and description

        Return:
            self
        """

        # If there is not any file passed as an argument, populate and empy table
        if not items_file:
            self.items = []
            return self

        # if there is, retrieve the content of the file to populate the table
        df = pd.read_csv(items_file, header=None)

        # TODO: We can check if the input file has header names, and if so, extract the
        # corresponding values with the SCHEMA.

        # small sanity check
        if not len(df.columns) in [2, 3]:
            raise AssertionError(
                f"The file is not a valid classification file as it has {len(df.columns)} columns instead of 2 or 3"
            )

        # add a color column if necessary
        if len(df.columns) == 2:
            df[2] = ["#000000"] * len(df)

        # TODO: Warning, here we are asumming that the values are in the same order as the schema keys
        # set the lines
        self.items = [
            dict(zip(self.SCHEMA.keys(), [i] + row.tolist()))
            for i, row in df.iterrows()
        ]

        return self

    def _save_event(self, widget, event, data):
        """open the save dialog to save the current table in a specific formatted table"""

        if not self.items:
            return

        self.save_dialog.v_model = True

        return

    def _edit_event(self, widget, event, data):
        """Open the edit dialog and fill it with current line information"""

        if not self.v_model:
            return

        self.edit_dialog.update([v for v in self.v_model[0].values()])

        return

    def _add_event(self, widget, event, data):
        """Open the edit dialog to create a new line to the table"""

        self.edit_dialog.update()

        return

    def _remove_event(self, widget, event, data):
        """Remove current selection (self.v_model) element from table"""

        if not self.v_model:
            return

        current_items = self.items.copy()
        current_items.remove(self.v_model[0])

        self.items = current_items

        return


class EditDialog(v.Dialog):
    """
    Dialog to modify/create new elements from the ClassTable data_table

    Args:
        table (ClassTable, v.DataTable): Table linked with dialog

    Attributes:
        TITLES (list): the list of potential header (translated at the start of the application)
        table (classTable): the classTable associated with the dialog
        title (v.CardTitle): the title of the card ('modify' or 'new')
        save (sw.Btn): the btn to validate the new line and save it in the datatable
        modify (sw.Btn): the btn to validate the edited line and save it in the datatable
        cancel (sw.Btn): the btn to close the dialog without saving anything
        widgets (list): the list of widget to control the new element state (id, code, description, color) in this order.
    """

    TITLES = ms.rec.table.edit_dialog.titles

    def __init__(self, table, **kwargs):

        # custom attributes
        self.table = table

        # set the title
        self.title = v.CardTitle(children=[self.TITLES[0]])

        # Action buttons
        btn_txt = ms.rec.table.edit_dialog.btn
        self.save = sw.Btn(btn_txt.save.name)
        save_tool = sw.Tooltip(self.save, btn_txt.save.tooltip, bottom=True)

        self.modify = sw.Btn(btn_txt.modify.name).hide()  # by default modify is hidden
        modify_tool = sw.Tooltip(self.modify, btn_txt.modify.tooltip, bottom=True)

        self.cancel = sw.Btn(btn_txt.cancel.name, outlined=True, class_="ml-2")
        cancel_tool = sw.Tooltip(self.cancel, btn_txt.cancel.tooltip, bottom=True)

        actions = v.CardActions(children=[save_tool, modify_tool, cancel_tool])

        # create the widgets
        self.widgets = [
            v.TextField(label=val[0], type=val[1], v_model=None, _metadata={"name": k})
            for k, val in param.SCHEMA.items()
            if k in ["id", "code", "desc"]
        ] + [
            v.ColorPicker(
                label=param.SCHEMA["color"][0],
                mode=param.SCHEMA["color"][1],
                _metadata={"name": "color"},
                v_model=None,
            )
        ]

        self.widgets[0].disabled = True  # it's the id

        # some default params
        self.v_model = False
        self.max_width = 500
        self.overlay_opcity = 0.7

        # create the object
        super().__init__(**kwargs)

        # add the widget to the dialog
        self.children = [
            v.Card(class_="pa-4", children=[self.title] + self.widgets + [actions])
        ]

        # Create events
        self.save.on_event("click", self._save)
        self.modify.on_event("click", self._modify)
        self.cancel.on_event("click", self._cancel)

    def update(self, data=[None, None, None, None]):
        """
        upadte the dialog with the provided information and activate it

        Args:
            data (list): the text value of the selected line (id, code, description, color). default to 4 None (new line)

        Return:
            self
        """

        # change the title accodring to the presence of data
        self.title.children = [self.TITLES[not any(data)]]

        # change the btns visiblity
        if not any(data):
            self.save.show()
            self.modify.hide()
        else:
            self.save.hide()
            self.modify.show()

        # change the content
        for i, item in enumerate(zip(self.widgets, data)):

            widget, val = item

            # textfield can be filled directly
            if i != 3:
                widget.v_model = val

            # the color picker need to be set in every color format to work
            else:

                # default to red if no value is provided
                val = val if val else "#FF0000"

                rgb = [int(i * 255) for i in to_rgb(val)]
                hls = rgb_to_hls(*to_rgb(val))
                hsv = rgb_to_hsv(*to_rgb(val))

                # rebuild all the colors from the hex code
                # the full v_model need to provided to see the color in the colorpicker
                widget.v_model = {
                    "alpha": 1,
                    "hex": val,
                    "hexa": f"{val}FF",
                    "hsla": {"h": hls[0], "s": hls[2], "l": hls[1], "a": 1},
                    "hsva": {"h": hsv[0], "s": hsv[1], "v": hsv[2], "a": 1},
                    "hue": hsv[0],
                    "rgba": {"r": rgb[0], "g": rgb[1], "b": rgb[2], "a": 1},
                }

        # if nothing is set I need to specify a new id value
        # this value should not interfere with the currently existing one. I'll thus just take the biggest +1
        if not any(data):
            self.widgets[0].v_model = (
                1
                if not self.table.items
                else max([i["id"] for i in self.table.items]) + 1
            )

        # activate the dialog
        self.v_model = True

        return self

    def _modify(self, widget, event, data):
        """Modify elements in the data_table and close the dialog"""

        # modify a local copy of the items
        # modifying does not trigger the display so a dummy element is also added
        # just to be removed afterward
        current_items = self.table.items.copy()
        for i, item in enumerate(current_items):
            if item["id"] == self.widgets[0].v_model:
                for w in self.widgets:
                    val = (
                        w.v_model
                        if w._metadata["name"] != "color"
                        else w.v_model["hex"]
                    )
                    current_items[i][w._metadata["name"]] = val

        current_items.append([""] * 4)

        # update the table values
        self.table.items = current_items

        # remove the dummy element to trigger the display
        self.table.items = self.table.items.copy()[:-1]

        # hide the dialog
        self.v_model = False

        # deselect the value for next time
        # if not deselected the previous item will remain in the v_model
        self.table.v_model = []

        return

    def _save(self, widget, event, data):
        """Add elements to the table and close the dialog"""

        # modify a local copy of the items
        current_items = self.table.items.copy()

        item_to_add = {}
        for w in self.widgets:
            item_to_add[w._metadata["name"]] = (
                w.v_model if w._metadata["name"] != "color" else w.v_model["hex"]
            )

        current_items.insert(0, item_to_add)

        # update the table values
        self.table.items = current_items

        # hide the dialog
        self.v_model = False

        return

    def _cancel(self, widget, event, data):
        """Close dialog and do nothing"""

        self.v_model = False

        return


class SaveDialog(v.Dialog):
    """
    Dialog to save as .csv file the content of a ClassTable data table

    Args:
        table (ClassTable): Table linked with dialog
        out_path (str): Folder path to store table content

    Attributes:
        reload (Int): a traitlet to inform the rest of the app that saving is complete
        table (ClassTable): Table linked with dialog
        out_path (str): Folder path to store table content
        w_file_name (v.TextField): the filename to use in out_path
        save (sw.Btn): save btn to launch the saving of the table data in the out_path using the filename provided in the widget
        cancel (sw.Btn): btn to close the dialog and do nothing
        alert (sw.Alert): an alert to display evolution of the saving process (errors)
    """

    reload = Int().tag(sync=True)

    def __init__(self, table, out_path, **kwargs):

        # gather the table and saving params
        self.table = table
        self.out_path = out_path

        # set some default parameters
        self.max_width = 500
        self.v_model = False

        # create the widget
        super().__init__(**kwargs)

        # build widgets
        self.w_file_name = v.TextField(
            label=ms.rec.table.save_dialog.filename,
            v_model=ms.rec.table.save_dialog.placeholder,
        )

        self.save = sw.Btn(ms.rec.table.save_dialog.btn.save.name)
        save = sw.Tooltip(
            self.save,
            ms.rec.table.save_dialog.btn.save.tooltip,
            bottom=True,
            class_="pr-2",
        )

        self.cancel = sw.Btn(
            ms.rec.table.save_dialog.btn.cancel.name, outlined=True, class_="ml-2"
        )
        cancel = sw.Tooltip(
            self.cancel, ms.rec.table.save_dialog.btn.cancel.tooltip, bottom=True
        )

        self.alert = sw.Alert(children=["Choose a name for the output"]).show()

        # assemlble the layout
        self.children = [
            v.Card(
                class_="pa-4",
                children=[
                    v.CardTitle(children=[ms.rec.table.save_dialog.title]),
                    self.w_file_name,
                    self.alert,
                    save,
                    cancel,
                ],
            )
        ]

        # Create events
        self.save.on_event("click", self._save)
        self.cancel.on_event("click", self._cancel)
        self.w_file_name.on_event("blur", self._normalize_name)
        self.w_file_name.observe(self._store_info, "v_model")

    def _store_info(self, change):
        """Display where will be the file written"""

        new_val = change["new"]
        out_file = self.out_path / f"{su.normalize_str(new_val)}.csv"

        msg = f"Your file will be saved as: {out_file}"

        if not new_val:
            msg = "Choose a name for the output"

        self.alert.add_msg(msg)

    def show(self):
        """
        display the dialog and write down the text in the alert

        Return:
            self
        """

        self.v_model = True
        self.w_file_name.v_model = ""

        # the message is display after the show so that it's not cut by the display
        self.alert.add_msg(ms.rec.table.save_dialog.info.format(self.out_path))

        return self

    def _normalize_name(self, widget, event, data):
        """Replace the name with it's normalized version"""

        # normalized the name
        widget.v_model = su.normalize_str(widget.v_model)

        return

    def _save(self, widget, event, data):
        """Write current table on a text file"""

        # set the file name
        out_file = self.out_path / su.normalize_str(self.w_file_name.v_model)

        # write each line values but not the id
        lines = [list(item.values())[1:] for item in self.table.items]
        txt = [",".join(str(e) for e in ln) + "\n" for ln in lines]
        out_file.with_suffix(".csv").write_text("".join(txt))

        # Every time a file is saved, we update the current widget state
        # so it can be observed by other objects.
        self.reload += 1

        self.v_model = False

        return

    def _cancel(self, widget, event, data):
        """hide the widget and do nothing"""

        self.v_model = False

        return


class TableView(sw.Card):
    """
    Stand-alone Card object allowing the user to build custom class table. The user can start from an existing table or start from scratch. It gives the oportunity to change: the value, the class name and the color. It can be used as a tile in a sepal_ui app. The id_ of the tile is set to "classification_tile"

    Args:
        class_path (str|optional): Folder path containing already existing classes. Default to ~/
        out_path (str|optional): the folder to save the created classifications. default to ~/downloads
    """

    title = None
    "v.CardTitle: the title of the card"

    class_path = None
    "str: Folder path containing already existing classes"

    out_path = None
    "str: the folder to save the created classifications"

    w_class_file = None
    "sw.FileInput: the file input of the existing classification system"

    w_class_table = None
    "ClassTable: the classtable (CRUD) to manage the editing of the classification"

    btn = None
    "sw.Btn: the btn to start loading file data into the table"

    alert = None
    "sw.Alert: the alert to display loading information (error and success)"

    def __init__(
        self, class_path=Path.home(), out_path=Path.home() / "downloads", **kwargs
    ):

        # create metadata to make it compatible with the framwork app system
        self._metadata = {"mount_id": "reclassify_tile"}

        # set some default params
        self.class_ = "pa-5"

        # init the card
        super().__init__(**kwargs)

        # set the folders
        self.class_path = Path(class_path)
        self.out_path = Path(out_path)

        # set a title to the card
        self.title = v.CardTitle(
            children=[v.Html(tag="h2", children=[ms.rec.table.title])]
        )

        # add the widgets
        w_class_title = v.Html(
            tag="h3", children=[ms.rec.table.classif.title], class_="mt-2"
        )
        self.w_class_file = sw.FileInput(
            extentions=[".csv"],
            label=ms.rec.table.classif.file_select,
            folder=self.class_path,
        )
        self.btn = sw.Btn(
            ms.rec.table.classif.btn,
            icon="far fa-table",
            color="success",
            outlined=True,
        )
        w_panels = v.ExpansionPanels(
            children=[
                v.ExpansionPanel(
                    children=[
                        v.ExpansionPanelHeader(children=[w_class_title]),
                        v.ExpansionPanelContent(children=[self.w_class_file, self.btn]),
                    ]
                )
            ]
        )

        w_table_title = v.Html(tag="h2", children=[ms.rec.table.table], class_="mt-5")
        self.w_class_table = ClassTable(out_path=self.out_path, class_="mt-5")

        # create an alert to display error and outputs
        self.alert = sw.Alert()

        # assemble a layout
        self.children = [
            self.title,
            w_panels,
            self.alert,
            w_table_title,
            self.w_class_table,
        ]

        # Events
        self.btn.on_event("click", self.get_class_table)

    @su.loading_button(debug=True)
    def get_class_table(self, widget, event, data):
        """
        Display class table widget in view

        Return:
            self
        """

        # load the existing file into the table
        self.w_class_table.populate_table(self.w_class_file.v_model)

        return self

    def nest_tile(self):
        """
        Prepare the view to be used as a nested component in a tile.
        the elevation will be set to 0 and the title remove from children.
        The mount_id will also be changed to nested

        Return:
            self
        """

        # remove id
        self._metadata["mount_id"] = "nested_tile"

        # remove elevation
        self.elevation = False

        # remove title
        without_title = self.children.copy()
        without_title.remove(self.title)
        self.children = without_title

        return self



import ipyvuetify as v
from traitlets import Unicode, Bool, observe

__all__ = ["TYPES", "SepalWidget"]

TYPES = ("info", "secondary", "primary", "error", "warning", "success", "accent")


class SepalWidget(v.VuetifyWidget):
    """
    Custom vuetifyWidget to add specific methods
    """

    viz = Bool(True).tag(sync=True)
    "Bool: whether the widget is displayed or not"

    old_class = Unicode("").tag(sync=True)
    "Unicode: a saving attribute of the widget class"

    def __init__(self, **kwargs):

        # remove viz from kwargs
        # class_list need to be setup before viz
        # to let hide and shw function run
        viz = kwargs.pop("viz", True)

        # init the widget
        super().__init__(**kwargs)

        # setup the viz status
        self.viz = viz

    @observe("viz")
    def _set_viz(self, change):
        """
        hide or show the component according to its viz param value.

        Hide the widget by reducing the html class to :code:`d-none`.
        Show the widget by removing the :code:`d-none` html class.
        Save the previous class

        Args:
            change: the dict of a trait callback
        """

        # will be replaced byt direct calls to built-in hide
        # once the previous custom implementation will be fully removed

        if self.viz:

            # change class value
            self.class_ = self.old_class or self.class_
            self.class_list.remove("d-none")

        else:

            # change class value
            self.class_list.remove("d-none")
            self.old_class = str(self.class_)
            self.class_ = "d-none"

        return

    def toggle_viz(self):
        """
        toogle the visibility of the widget.

        Return:
            self
        """

        self.viz = not self.viz

        return self

    def hide(self):
        """
        Hide the widget by reducing the html class to :code:`d-none`.
        Save the previous class and set viz attribute to False.

        Return:
            self
        """

        # update viz state
        self.viz = False

        return self

    def show(self):
        """
        Show the widget by removing the d-none html class.
        Save the previous class and set viz attribute to True.

        Return:
            self
        """

        # update viz state
        self.viz = True

        return self

    def reset(self):
        """
        Clear the widget v_model. Need to be extented in custom widgets to fit the structure of the actual input.

        Return:
            self
        """

        self.v_model = None

        return self



from pathlib import Path

import ipyvuetify as v
from traitlets import link, Int, Any, List, observe, Dict, Unicode
from ipywidgets import jslink
import pandas as pd
import ee
import geopandas as gpd
from natsort import humansorted


from sepal_ui.message import ms
from sepal_ui.frontend.styles import COMPONENTS, ICON_TYPES
from sepal_ui.scripts import utils as su
from sepal_ui.scripts import gee
from sepal_ui.sepalwidgets.sepalwidget import SepalWidget
from sepal_ui.sepalwidgets.btn import Btn

__all__ = [
    "DatePicker",
    "FileInput",
    "LoadTableField",
    "AssetSelect",
    "PasswordField",
    "NumberField",
    "VectorField",
]


class DatePicker(v.Layout, SepalWidget):
    """
    Custom input widget to provide a reusable DatePicker. It allows to choose date as a string in the following format YYYY-MM-DD

    Args:
        label (str, optional): the label of the datepicker field
        kwargs (optional): any parameter from a v.Layout abject. If set, 'children' will be overwritten.

    """

    menu = None
    "v.Menu: the menu widget to display the datepicker"

    def __init__(self, label="Date", **kwargs):

        # create the widgets
        date_picker = v.DatePicker(no_title=True, v_model=None, scrollable=True)

        date_text = v.TextField(
            v_model=None,
            label=label,
            hint="YYYY-MM-DD format",
            persistent_hint=True,
            prepend_icon="event",
            readonly=True,
            v_on="menuData.on",
        )

        self.menu = v.Menu(
            min_width="290px",
            transition="scale-transition",
            offset_y=True,
            v_model=False,
            close_on_content_click=False,
            children=[date_picker],
            v_slots=[
                {
                    "name": "activator",
                    "variable": "menuData",
                    "children": date_text,
                }
            ],
        )

        # set the default parameter
        kwargs["v_model"] = kwargs.pop("v_model", None)
        kwargs["row"] = kwargs.pop("row", True)
        kwargs["class_"] = kwargs.pop("class_", "pa-5")
        kwargs["align_center"] = kwargs.pop("align_center", True)
        kwargs["children"] = [v.Flex(xs10=True, children=[self.menu])]

        # call the constructor
        super().__init__(**kwargs)

        jslink((date_picker, "v_model"), (date_text, "v_model"))
        jslink((date_picker, "v_model"), (self, "v_model"))

    @observe("v_model")
    def close_menu(self, change):
        """A method to close the menu of the datepicker programatically"""

        # set the visibility
        self.menu.v_model = False

        return


class FileInput(v.Flex, SepalWidget):
    """
    Custom input field to select a file in the sepal folders.

    Args:
        extentions ([str]): the list of the allowed extentions. the FileInput will only display these extention and folders
        folder (str | pathlib.Path): the starting folder of the file input
        label (str): the label of the input
        v_model (str, optional): the default value
        clearable (bool, optional): wether or not to make the widget clearable. default to False
        kwargs (optional): any parameter from a v.Flex abject. If set, 'children' will be overwritten.
    """

    extentions = []
    "list: the extention list"

    folder = Path.home()
    "pathlib.Path: the current folder"

    file = Any("").tag(sync=True)
    "str: the current file"

    selected_file = None
    "v.TextField: the textfield where the file pathname is stored"

    loading = None
    "v.ProgressLinear: loading top bar of the menu component"

    file_list = None
    "v.List: the list of files and folder that are available in the current folder"

    file_menu = None
    "v.Menu: the menu that hide and show the file_list"

    reload = None
    "v.Btn: reload btn to reload the file list on the current folder"

    clear = None
    "v.Btn: clear btn to remove everything and set back to the ini folder"

    v_model = Unicode(None, allow_none=True).tag(sync=True)
    "str: the v_model of the input"

    def __init__(
        self,
        extentions=[],
        folder=Path.home(),
        label="search file",
        v_model=None,
        clearable=False,
        **kwargs,
    ):

        if type(folder) == str:
            folder = Path(folder)

        self.extentions = extentions
        self.folder = folder

        self.selected_file = v.TextField(
            readonly=True, label="Selected file", class_="ml-5 mt-5", v_model=None
        )

        self.loading = v.ProgressLinear(
            indeterminate=False,
            background_color="grey darken-3",
            color=COMPONENTS["PROGRESS_BAR"]["color"],
        )

        self.file_list = v.List(
            dense=True,
            color="grey darken-3",
            flat=True,
            v_model=True,
            max_height="300px",
            style_="overflow: auto; border-radius: 0 0 0 0;",
            children=[v.ListItemGroup(children=self._get_items(), v_model="")],
        )

        self.file_menu = v.Menu(
            min_width=300,
            children=[self.loading, self.file_list],
            v_model=False,
            close_on_content_click=False,
            v_slots=[
                {
                    "name": "activator",
                    "variable": "x",
                    "children": Btn(
                        icon="fas fa-search", v_model=False, v_on="x.on", text=label
                    ),
                }
            ],
        )

        self.reload = v.Btn(
            icon=True, color="primary", children=[v.Icon(children=["fas fa-sync-alt"])]
        )

        self.clear = v.Btn(
            icon=True, color="primary", children=[v.Icon(children=["fas fa-times"])]
        )
        if not clearable:
            su.hide_component(self.clear)

        # set default parameters
        kwargs["row"] = kwargs.pop("row", True)
        kwargs["class_"] = kwargs.pop("class_", "d-flex align-center mb-2")
        kwargs["align_center"] = kwargs.pop("align_center", True)
        kwargs["children"] = [
            self.clear,
            self.reload,
            self.file_menu,
            self.selected_file,
        ]

        # call the constructor
        super().__init__(**kwargs)

        link((self.selected_file, "v_model"), (self, "file"))
        link((self.selected_file, "v_model"), (self, "v_model"))

        self.file_list.children[0].observe(self._on_file_select, "v_model")
        self.reload.on_event("click", self._on_reload)
        self.clear.on_event("click", self.reset)

    def reset(self, *args):
        """
        Clear the File selection and move to the root folder if something was selected

        Return:
            self
        """

        root = Path("~").expanduser()

        if self.v_model is not None:

            # move to root
            self._on_file_select({"new": root})

            # remove v_model
            self.v_model = None

        return self

    def select_file(self, path):
        """
        Manually select a file from it's path. No verification on the extension is performed

        Params:
            path (str|pathlib.Path): the path to the file

        Return:
            self
        """

        # cast to Path
        path = Path(path)

        # test file existence
        if not path.is_file():
            raise Exception(f"{path} is not a file")

        # set the menu to the folder of the file
        self._on_file_select({"new": path.parent})

        # select the appropriate file
        self._on_file_select({"new": path})

        return self

    def _on_file_select(self, change):
        """Dispatch the behavior between file selection and folder change"""

        if not change["new"]:
            return self

        new_value = Path(change["new"])

        if new_value.is_dir():
            self.folder = new_value
            self._change_folder()

        elif new_value.is_file():
            self.file = str(new_value)

        return self

    @su.switch("indeterminate", on_widgets=["loading"])
    def _change_folder(self):
        """Change the target folder"""

        # get the items
        items = self._get_items()

        # reset files
        # this is reseting the scroll to top without using js scripts
        self.file_list.children[0].children = []

        # set the new files
        self.file_list.children[0].children = items

        return

    def _get_items(self):
        """Return the list of items inside the folder"""

        folder = self.folder

        list_dir = [el for el in folder.glob("*/") if not el.name.startswith(".")]

        if self.extentions:
            list_dir = [
                el for el in list_dir if el.is_dir() or el.suffix in self.extentions
            ]

        folder_list = []
        file_list = []

        for el in list_dir:

            if el.is_dir():
                icon = ICON_TYPES[""]["icon"]
                color = ICON_TYPES[""]["color"]
            elif el.suffix in ICON_TYPES.keys():
                icon = ICON_TYPES[el.suffix]["icon"]
                color = ICON_TYPES[el.suffix]["color"]
            else:
                icon = ICON_TYPES["DEFAULT"]["icon"]
                color = ICON_TYPES["DEFAULT"]["color"]

            children = [
                v.ListItemAction(children=[v.Icon(color=color, children=[icon])]),
                v.ListItemContent(
                    children=[v.ListItemTitle(children=[el.stem + el.suffix])]
                ),
            ]

            if el.is_dir():
                folder_list.append(v.ListItem(value=str(el), children=children))
            else:
                file_size = su.get_file_size(el)
                children.append(v.ListItemActionText(children=[file_size]))
                file_list.append(v.ListItem(value=str(el), children=children))

        folder_list = humansorted(folder_list, key=lambda x: x.value)
        file_list = humansorted(file_list, key=lambda x: x.value)

        parent_item = v.ListItem(
            value=str(folder.parent),
            children=[
                v.ListItemAction(
                    children=[
                        v.Icon(
                            color=ICON_TYPES["PARENT"]["color"],
                            children=[ICON_TYPES["PARENT"]["icon"]],
                        )
                    ]
                ),
                v.ListItemContent(
                    children=[v.ListItemTitle(children=[f"..{folder.parent}"])]
                ),
            ],
        )

        folder_list.extend(file_list)
        folder_list.insert(0, parent_item)

        return folder_list

    def _on_reload(self, widget, event, data):

        # force the update of the current folder
        self._change_folder()

        return

    @observe("v_model")
    def close_menu(self, change):
        """A method to close the menu of the Fileinput programatically"""

        # set the visibility
        self.file_menu.v_model = False

        return


class LoadTableField(v.Col, SepalWidget):
    """
    A custom input widget to load points data. The user will provide a csv or txt file containing labeled dataset.
    The relevant columns (lat, long and id) can then be identified in the updated select. Once everything is set, the widget will populate itself with a json dict.
    {pathname, id_column, lat_column,lng_column}

    Args:
        label (str, optional): the label of the widget
        kwargs (optional): any parameter from a v.Col. If set, 'children' and 'v_model' will be overwritten.
    """

    fileInput = None
    "sw.FileInput: the file input to select the .csv or .txt file"

    IdSelect = None
    "v.Select: input to select the id column"

    LngSelect = None
    "v.Select: input to select the lng column"

    LatSelect = None
    "v.Select: input to select the lat column"

    default_v_model = {
        "pathname": None,
        "id_column": None,
        "lat_column": None,
        "lng_column": None,
    }
    "dict: The default v_model structure {'pathname': xx, 'id_column': xx, 'lat_column': xx, 'lng_column': xx}"

    def __init__(self, label="Table file", **kwargs):

        self.fileInput = FileInput([".csv", ".txt"], label=label)

        self.IdSelect = v.Select(
            _metadata={"name": "id_column"}, items=[], label="Id", v_model=None
        )
        self.LngSelect = v.Select(
            _metadata={"name": "lng_column"}, items=[], label="Longitude", v_model=None
        )
        self.LatSelect = v.Select(
            _metadata={"name": "lat_column"}, items=[], label="Latitude", v_model=None
        )

        # set default parameters
        kwargs["v_model"] = self.default_v_model  # format of v_model is fixed
        kwargs["children"] = [
            self.fileInput,
            self.IdSelect,
            self.LngSelect,
            self.LatSelect,
        ]

        # call the constructor
        super().__init__(**kwargs)

        # link the dropdowns
        jslink((self.IdSelect, "items"), (self.LngSelect, "items"))
        jslink((self.IdSelect, "items"), (self.LatSelect, "items"))

        # link the widget with v_model
        self.fileInput.observe(self._on_file_input_change, "v_model")
        self.IdSelect.observe(self._on_select_change, "v_model")
        self.LngSelect.observe(self._on_select_change, "v_model")
        self.LatSelect.observe(self._on_select_change, "v_model")

    def reset(self):
        """
        Clear the values and return to the empty default json

        Return:
            self
        """

        # clear the fileInput
        self.fileInput.reset()

    @su.switch("loading", on_widgets=["IdSelect", "LngSelect", "LatSelect"])
    def _on_file_input_change(self, change):
        """Update the select content when the fileinput v_model is changing"""

        # clear the selects
        self._clear_select()

        # set the path
        path = change["new"]
        self.v_model["pathname"] = path

        # exit if none
        if not path:
            return self

        df = pd.read_csv(path, sep=None, engine="python")

        if len(df.columns) < 3:
            self._clear_select()
            self.fileInput.selected_file.error_messages = (
                ms.widgets.load_table.too_small
            )
            return

        # set the items
        self.IdSelect.items = df.columns.tolist()

        # pre load values that sounds like what we are looking for
        # it will only keep the first occurence of each one
        for name in reversed(df.columns.tolist()):
            lname = name.lower()
            if "id" in lname:
                self.IdSelect.v_model = name
            elif any(
                ext in lname
                for ext in ["lng", "long", "longitude", "x_coord", "xcoord", "lon"]
            ):
                self.LngSelect.v_model = name
            elif any(ext in lname for ext in ["lat", "latitude", "y_coord", "ycoord"]):
                self.LatSelect.v_model = name

        return self

    def _clear_select(self):
        """clear the selects components"""

        self.fileInput.selected_file.error_messages = None
        self.IdSelect.items = []  # all the others are listening to this one
        self.IdSelect.v_model = self.LngSelect.v_model = self.LatSelect.v_model = None

        return self

    def _on_select_change(self, change):
        """change the v_model value when a select is changed"""

        name = change["owner"]._metadata["name"]
        self.v_model[name] = change["new"]

        return self


class AssetSelect(v.Combobox, SepalWidget):
    """
    Custom widget input to select an asset inside the asset folder of the user

    Args:
        label (str): the label of the input
        folder (str): the folder of the user assets
        default_asset (str, list): the id of a default asset or a list of defaults
        types ([str]): the list of asset type you want to display to the user. type need to be from: ['IMAGE', 'FOLDER', 'IMAGE_COLLECTION', 'TABLE','ALGORITHM']. Default to 'IMAGE' & 'TABLE'
        kwargs (optional): any parameter from a v.ComboBox.
    """

    TYPES = {
        "IMAGE": ms.widgets.asset_select.types[0],
        "TABLE": ms.widgets.asset_select.types[1],
        "IMAGE_COLLECTION": ms.widgets.asset_select.types[2],
        "ALGORITHM": ms.widgets.asset_select.types[3],
        "FOLDER": ms.widgets.asset_select.types[4],
        # UNKNOWN type is ignored
    }
    "dict: Valid ypes of asset"

    folder = None
    "str: the folder of the user assets, mainly for debug"

    valid = True
    "Bool: whether the selected asset is valid (user has access) or not"

    asset_info = {}
    "dict: The selected asset informations"

    default_asset = Any().tag(sync=True)
    "str: the id of a default asset or a list of default assets"

    types = List().tag(sync=True)
    "List: the list of types accepted by the asset selector. names need to be valide TYPES and changing this value will trigger the reload of the asset items."

    @su.need_ee
    def __init__(
        self,
        label=ms.widgets.asset_select.label,
        folder=None,
        types=["IMAGE", "TABLE"],
        **kwargs,
    ):
        self.valid = False
        self.asset_info = None

        # if folder is not set use the root one
        self.folder = folder if folder else ee.data.getAssetRoots()[0]["id"]
        self.types = types

        # Validate the input as soon as the object is instantiated
        self.observe(self._validate, "v_model")

        # load the assets in the combobox
        self._get_items()

        # set the default parameters
        kwargs["v_model"] = kwargs.pop("v_model", None)
        kwargs["clearable"] = kwargs.pop("clearable", True)
        kwargs["dense"] = kwargs.pop("dense", True)
        kwargs["prepend_icon"] = kwargs.pop("prepend_icon", "fas fa-sync-alt")
        kwargs["class_"] = kwargs.pop("class_", "my-5")
        kwargs["placeholder"] = kwargs.pop(
            "placeholder", ms.widgets.asset_select.placeholder
        )

        # create the widget
        super().__init__(**kwargs)

        # add js behaviours
        self.on_event("click:prepend", self._get_items)

    @observe("default_asset")
    def _add_default(self, change=None):
        """Add default element(s) to item list"""

        new_val = change["new"]

        if new_val:

            new_items = self.items.copy()

            if isinstance(new_val, str):
                new_val = [new_val]

            self.v_model = new_val[0]

            default_items = [
                {"divider": True},
                {"header": ms.widgets.asset_select.custom},
            ] + [default for default in new_val]

            # Check if there are previously custom items and replace them, instead
            # of append them

            if {"header": ms.widgets.asset_select.custom} in self.items:

                next_div_index = new_items.index({"divider": True}, 1)
                new_items[0:next_div_index] = default_items

            else:
                new_items = default_items + self.items

            self.items = new_items

        return

    @su.switch("loading")
    def _validate(self, change):
        """
        Validate the selected access. Throw an error message if is not accesible or not in the type list.
        """

        self.error_messages = None

        if change["new"]:

            # check that the asset can be accessed
            try:
                self.asset_info = ee.data.getAsset(change["new"])

                # check that the asset has the correct type
                if not self.asset_info["type"] in self.types:
                    self.error_messages = ms.widgets.asset_select.wrong_type.format(
                        self.asset_info["type"], ",".join(self.types)
                    )

            except Exception:

                self.error_messages = ms.widgets.asset_select.no_access

            self.valid = self.error_messages is None
            self.error = self.error_messages is not None

        return

    @su.switch("loading", "disabled")
    def _get_items(self, change=None):

        # get the list of user asset
        raw_assets = gee.get_assets(self.folder)

        assets = {
            k: sorted([e["id"] for e in raw_assets if e["type"] == k])
            for k in self.types
        }

        items = []
        for k in self.types:
            if len(assets[k]):
                items += [
                    {"divider": True},
                    {"header": self.TYPES[k]},
                    *assets[k],
                ]

        self.items = items

        return self

    @observe("types")
    def _check_types(self, change):
        """clean the type list, keeping only the valid one"""

        self.v_model = None

        # check the type
        self.types = [t for t in self.types if t in self.TYPES]

        # trigger the reload
        self._get_items()

        return


class PasswordField(v.TextField, SepalWidget):
    """
    Custom widget to input passwords in text area and
    toggle its visibility.

    Args:
        label (str, optional): Header displayed in text area. Defaults to Password.
        kwargs (dict); any parameter from a v.TextField. If set, 'type' will be overwritten.
    """

    def __init__(self, **kwargs):

        # default behavior
        kwargs["label"] = kwargs.pop("label", "Password")
        kwargs["class_"] = kwargs.pop("class_", "mr-2")
        kwargs["v_model"] = kwargs.pop("v_model", "")
        kwargs["type"] = "password"
        kwargs["append_icon"] = kwargs.pop("append_icon", "far fa-eye-slash")

        # init the widget with the remaining kwargs
        super().__init__(**kwargs)

        # bind the js behavior
        self.on_event("click:append", self._toggle_pwd)

    def _toggle_pwd(self, widget, event, data):
        """Toggle password visibility when append button is clicked"""

        if self.type == "text":
            self.type = "password"
            self.append_icon = "far fa-eye-slash"
        else:
            self.type = "text"
            self.append_icon = "far fa-eye"


class NumberField(v.TextField, SepalWidget):
    """
    Custom widget to input numbers in text area and add/substract with single increment.

    Args:
        max_ (int, optional): Maximum selectable number. Defaults to 10.
        min_ (int, optional): Minimum selectable number. Defaults to 0.
        increm (int, optional): incremental value added at each step. default to 1
        kwargs (dict, optional): Any parameter from a v.TextField. If set, 'type' will be overwritten.
    """

    max_ = Int(10).tag(sync=True)
    "int: Maximum selectable number."

    min_ = Int(0).tag(sync=True)
    "int: Minimum selectable number."

    increm = Int(1).tag(sync=True)
    "int: incremental value added at each step."

    def __init__(self, max_=10, min_=0, increm=1, **kwargs):

        # set the traits
        self.max_ = max_
        self.min_ = min_
        self.increm = increm

        # set default params
        kwargs["type"] = "number"
        kwargs["append_outer_icon"] = kwargs.pop("append_outer_icon", "fas fa-plus")
        kwargs["prepend_icon"] = kwargs.pop("prepend_icon", "fas fa-minus")
        kwargs["v_model"] = kwargs.pop("v_model", 0)
        kwargs["readonly"] = kwargs.pop("readonly", True)

        # call the constructor
        super().__init__(**kwargs)

        self.on_event("click:append-outer", self.increment)
        self.on_event("click:prepend", self.decrement)

    def increment(self, widget, event, data):
        """Adds increm to the current v_model number"""

        self.v_model = min((self.v_model + self.increm), self.max_)

        return

    def decrement(self, widget, event, data):
        """Substracts increm to the current v_model number"""

        self.v_model = max((self.v_model - self.increm), self.min_)

        return


class VectorField(v.Col, SepalWidget):
    """
    A custom input widget to load vector data. The user will provide a vector file compatible with fiona or a GEE feature collection.
    The user can then select a specific shape by setting column and value fields.

    Args:
        label (str): the label of the file input field, default to 'vector file'.
        gee (bool, optional): whether to use GEE assets or local vectors.
        folder (str, optional): When gee=True, extra args will be used for AssetSelect
        kwargs (dict, optional): any parameter from a v.Col. if set, 'children' will be overwritten.
    """

    original_gdf = None
    "geopandas.GeoDataframe: The originally selected dataframe"

    df = None
    "pandas.Dataframe: the orginal dataframe without the geometry (for column naming)"

    gdf = None
    "geopandas.GeoDataframe: The selected dataframe"

    w_file = None
    "sw.FileInput: The file selector widget"

    w_column = None
    "v.Select: The Select widget to select the column"

    w_value = None
    "v.Select: The Select widget to select the value in the selected column"

    v_model = Dict(
        {
            "pathname": None,
            "column": None,
            "value": None,
        }
    )
    "Traitlet: The json saved v_model shaped as {'pathname': xx, 'column': xx, 'value': xx}"

    column_base_items = [
        {"text": "Use all features", "value": "ALL"},
        {"divider": True},
    ]
    "list: the column compulsory selector (ALL)"

    feature_collection = None
    "ee.FeatureCollection: the selected featureCollection"

    def __init__(self, label="vector_file", gee=False, **kwargs):

        # set the 3 wigets
        if not gee:
            self.w_file = FileInput([".shp", ".geojson", ".gpkg", ".kml"], label=label)
        else:
            # Don't care about 'types' arg. It will only work with tables.
            asset_select_kwargs = {"folder": kwargs.pop("folder", None)}
            self.w_file = AssetSelect(types=["TABLE"], **asset_select_kwargs)

        self.w_column = v.Select(
            _metadata={"name": "column"},
            items=self.column_base_items,
            label="Column",
            v_model="ALL",
        )
        self.w_value = v.Select(
            _metadata={"name": "value"}, items=[], label="Value", v_model=None
        )
        su.hide_component(self.w_value)

        # create the Col Field
        kwargs["children"] = [self.w_file, self.w_column, self.w_value]

        super().__init__(**kwargs)

        # events
        self.w_file.observe(self._update_file, "v_model")
        self.w_column.observe(self._update_column, "v_model")
        self.w_value.observe(self._update_value, "v_model")

    def reset(self):
        """
        Return the field to its initial state

        Return:
            self
        """

        self.w_file.reset()

        return self

    @su.switch("loading", on_widgets=["w_column", "w_value"])
    def _update_file(self, change):
        """update the file name, the v_model and reset the other widgets"""

        # reset the widgets
        self.w_column.items, self.w_value.items = [], []
        self.w_column.v_model = self.w_value.v_model = None
        self.df = None
        self.feature_collection = None

        # set the pathname value
        self.v_model["pathname"] = change["new"]

        # exit if nothing
        if not change["new"]:
            return self

        if isinstance(self.w_file, FileInput):
            # read the file
            self.df = gpd.read_file(change["new"], ignore_geometry=True)
            columns = self.df.columns.to_list()

        elif isinstance(self.w_file, AssetSelect):
            self.feature_collection = ee.FeatureCollection(change["new"])
            columns = self.feature_collection.first().getInfo()["properties"]
            columns = [
                str(col) for col in columns if col not in ["system:index", "Shape_Area"]
            ]

        # update the columns
        self.w_column.items = self.column_base_items + sorted(set(columns))

        self.w_column.v_model = "ALL"

        return self

    @su.switch("loading", on_widgets=["w_value"])
    def _update_column(self, change):
        """Update the column name and empty the value list"""

        # reset value widget
        self.w_value.items = []
        self.w_value.v_model = None

        # set the value
        self.v_model["column"] = change["new"]

        # hide value if "ALL" or none
        if change["new"] in ["ALL", None]:
            su.hide_component(self.w_value)
            return self

        # read the colmun
        if isinstance(self.w_file, FileInput):
            values = self.df[change["new"]].to_list()

        elif isinstance(self.w_file, AssetSelect):
            values = (
                self.feature_collection.distinct(change["new"])
                .aggregate_array(change["new"])
                .getInfo()
            )

        self.w_value.items = sorted(set(values))

        su.show_component(self.w_value)

        return self

    def _update_value(self, change):
        """Update the value name and reduce the gdf"""

        # set the value
        self.v_model["value"] = change["new"]

        return self



from functools import partial
from datetime import datetime

import ipyvuetify as v
from deprecated.sphinx import versionadded

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget
from sepal_ui import color
from sepal_ui.frontend.styles import sepal_main, sepal_darker
from sepal_ui.frontend import js

__all__ = ["AppBar", "DrawerItem", "NavDrawer", "Footer", "App"]


class AppBar(v.AppBar, SepalWidget):
    """
    Custom AppBar widget with the provided title using the sepal color framework

    Args:
        title (str, optional): the title of the app
        kwargs(dict, optional): any parameters from a v.AppBar. If set, 'children' and 'app' will be overwritten.
    """

    toogle_button = None
    "v.Btn: The btn to display or hide the drawer to the user"

    title = None
    "v.ToolBarTitle: the widget containing the app title"

    def __init__(self, title="SEPAL module", **kwargs):

        self.toggle_button = v.Btn(
            icon=True,
            children=[v.Icon(class_="white--text", children=["fas fa-ellipsis-v"])],
        )

        self.title = v.ToolbarTitle(children=[title])

        # set the default parameters
        kwargs["color"] = kwargs.pop("color", sepal_main)
        kwargs["class_"] = kwargs.pop("class_", "white--text")
        kwargs["dense"] = kwargs.pop("dense", True)
        kwargs["app"] = True
        kwargs["children"] = [self.toggle_button, self.title]

        super().__init__(**kwargs)

    def set_title(self, title):
        """
        Set the title of the appBar

        Args:
            title (str): the new app title

        Return:
            self
        """

        self.title.children = [title]

        return self


class DrawerItem(v.ListItem, SepalWidget):
    """
    Custom DrawerItem using the user input.
    If a card is set the drawerItem will trigger the display of all the Tiles in the app that have the same mount_id.
    If an href is set, the drawer will open the link in a new tab

    Args:
        title (str): the title of the drawer item
        icon(str, optional): the full name of a mdi/fa icon
        card (str, optional): the mount_id of tiles in the app
        href (str, optional): the absolute link to an external web page
        kwargs (optional): any parameter from a v.ListItem. If set, '_metadata', 'target', 'link' and 'children' will be overwritten.
    """

    rt = None
    "sw.ResizeTrigger: the trigger to resize maps and other javascript object when jumping from a tile to another"

    def __init__(self, title, icon=None, card=None, href=None, **kwargs):

        # set the resizetrigger
        self.rt = js.rt

        icon = icon if icon else "far fa-folder"

        children = [
            v.ListItemAction(children=[v.Icon(class_="white--text", children=[icon])]),
            v.ListItemContent(
                children=[v.ListItemTitle(class_="white--text", children=[title])]
            ),
        ]

        # set default parameters
        kwargs["link"] = True
        kwargs["children"] = children
        kwargs["input_value"] = kwargs.pop("input_value", False)
        if href:
            kwargs["href"] = href  # cannot be set twice anyway
            kwargs["target"] = "_blank"
            kwargs["_metadata"] = kwargs.pop("_metadata", None)
        elif card:
            kwargs["_metadata"] = {"card_id": card}
            kwargs["href"] = kwargs.pop("href", None)
            kwargs["target"] = kwargs.pop("target", None)

        # call the constructor
        super().__init__(**kwargs)

    def display_tile(self, tiles):
        """
        Display the apropriate tiles when the item is clicked.
        The tile to display will be all tile in the list with the mount_id as the current object

        Args:
            tiles ([sw.Tile]) : the list of all the available tiles in the app

        Return:
            self
        """

        self.on_event("click", partial(self._on_click, tiles=tiles))

        return self

    def _on_click(self, widget, event, data, tiles):

        for tile in tiles:
            if self._metadata["card_id"] == tile._metadata["mount_id"]:
                tile.show()
            else:
                tile.hide()

        # trigger the resize event
        self.rt.resize()

        # change the current item status
        self.input_value = True

        return self


class NavDrawer(v.NavigationDrawer, SepalWidget):
    """
    Custom NavDrawer using the different DrawerItems of the user and the sepal color framework.
    The drawer can include links to the github page of the project for wiki, bugs and repository.

    Args:
        item ([sw.DrawerItem]): the list of all the drawerItem to display in the drawer. This items should pilote the different tile visibility
        code (str, optional): the absolute link to the source code
        wiki (str, optional): the absolute link the the wiki page
        issue (str, optional): the absolute link to the issue tracker
        kwargs (optional) any parameter from a v.NavigationDrawer. If set, 'app' and 'children' will be overwritten.
    """

    items = []
    "list: the list of all the drawerItem to display in the drawer"

    def __init__(self, items=[], code=None, wiki=None, issue=None, **kwargs):

        self.items = items

        code_link = []
        if code:
            item_code = DrawerItem("Source code", icon="far fa-file-code", href=code)
            code_link.append(item_code)
        if wiki:
            item_wiki = DrawerItem("Wiki", icon="fas fa-book-open", href=wiki)
            code_link.append(item_wiki)
        if issue:
            item_bug = DrawerItem("Bug report", icon="fas fa-bug", href=issue)
            code_link.append(item_bug)

        children = [
            v.List(dense=True, children=self.items),
            v.Divider(),
            v.List(dense=True, children=code_link),
        ]

        # set default parameters
        kwargs["v_model"] = kwargs.pop("v_model", True)
        kwargs["app"] = True
        kwargs["color"] = kwargs.pop("color", sepal_darker)
        kwargs["children"] = children

        # call the constructor
        super().__init__(**kwargs)

        # bind the javascripts behavior
        for i in self.items:
            i.observe(self._on_item_click, "input_value")

    def display_drawer(self, toggleButton):
        """
        Bind the drawer to the app toggleButton

        Args:
            toggleButton(v.Btn) : the button that activate the drawer
        """

        toggleButton.on_event("click", self._on_drawer_click)

        return self

    def _on_drawer_click(self, widget, event, data):
        """
        Toggle the drawer visibility
        """

        self.v_model = not self.v_model

        return self

    def _on_item_click(self, change):
        """
        Deactivate all the other items when on of the is activated
        """
        if change["new"] is False:
            return self

        # reset all others states
        for i in self.items:
            if i != change["owner"]:
                i.input_value = False

        return self


class Footer(v.Footer, SepalWidget):
    """
    Custom Footer with cuzomizable text.
    Not yet capable of displaying logos

    Args:
        text (str, optional): the text to display in the future
        kwargs (optional): any parameter from a v.Footer. If set ['app', 'children'] will be overwritten.
    """

    def __init__(self, text="", **kwargs):

        text = text if text != "" else "SEPAL \u00A9 {}".format(datetime.today().year)

        # set default parameters
        kwargs["color"] = kwargs.pop("color", sepal_main)
        kwargs["class_"] = kwargs.pop("class_", "white--text")
        kwargs["app"] = True
        kwargs["children"] = [text]

        # call the constructor
        super().__init__(**kwargs)


class App(v.App, SepalWidget):
    """
    Custom App display with the tiles created by the user using the sepal color framework.
    Display false appBar if not filled. Navdrawer is fully optionnal.
    The drawerItem will be linked to the app tile and they will be able to control their display
    If the navdrawer exist, it will be linked to the appbar togglebtn

    Args:
        tiles ([sw.Tile]): the tiles of the app
        appBar (sw.AppBar, optional): the appBar of the application
        footer (sw.Footer, optional): the footer of the application
        navDrawer (sw.NavDrawer): the navdrawer of the application
        kwargs (optional) any parameter from a v.App. If set, 'children' will be overwritten.
    """

    tiles = []
    "list: the tiles of the app"

    appBar = None
    "sw.AppBar: the AppBar of the application"

    footer = None
    "sw.Footer: the footer of the application"

    navDrawer = None
    "sw.NavDrawer: the navdrawer of the application"

    content = None
    "v.Content: the tiles organized in a fluid container"

    def __init__(self, tiles=[""], appBar=None, footer=None, navDrawer=None, **kwargs):

        self.tiles = None if tiles == [""] else tiles

        app_children = []

        # create a false appBar if necessary
        if not appBar:
            appBar = AppBar()
        self.appBar = appBar
        app_children.append(self.appBar)

        # add the navDrawer if existing
        self.navDrawer = None
        if navDrawer:
            # bind app tile list to the navdrawer
            for di in navDrawer.items:
                di.display_tile(tiles)

            # link it with the appbar
            navDrawer.display_drawer(self.appBar.toggle_button)

            # add the drawers to the children
            self.navDrawer = navDrawer
            app_children.append(self.navDrawer)

        # add the content of the app
        self.content = v.Content(children=[v.Container(fluid=True, children=tiles)])
        app_children.append(self.content)

        # add the footer if existing
        if footer:
            self.footer = footer
            app_children.append(self.footer)

        # create a negative overlay to force the background color
        bg = v.Overlay(color=color.bg, opacity=1, style_="transition:unset", z_index=-1)

        # set default parameters
        kwargs["v_model"] = kwargs.pop("v_model", None)
        kwargs["children"] = [bg, *app_children]

        # call the constructor
        super().__init__(**kwargs)

    def show_tile(self, name):
        """
        Select the tile to display when the app is launched

        Args:
            name (str): the mount-id of the tile(s) to display

        Return:
            self
        """
        # show the tile
        for tile in self.tiles:
            if name == tile._metadata["mount_id"]:
                tile.show()
            else:
                tile.hide()

        # activate the drawerItem
        if self.navDrawer:
            items = (i for i in self.navDrawer.items if i._metadata is not None)
            for i in items:
                if name == i._metadata["card_id"]:
                    i.input_value = True

        return self

    @versionadded(version="2.4.1", reason="New end user interaction method")
    def add_banner(self, msg, **kwargs):
        """
        Display an alert object on top of the app to communicate development information to end user (release date, known issues, beta version). The alert is dissmisable and prominent

        Args:
            msg (str): the message to write in the Alert
            kwargs: any arguments of the v.Alert constructor. if set, 'children' will be overwritten.

        Return:
            self
        """

        kwargs["type"] = kwargs.pop("type", "info")
        kwargs["border"] = kwargs.pop("border", "left")
        kwargs["class_"] = kwargs.pop("class_", "mt-5")
        kwargs["transition"] = kwargs.pop("transition", "slide-x-transition")
        kwargs["prominent"] = kwargs.pop("prominent", True)
        kwargs["dismissible"] = kwargs.pop("dismissible", True)
        kwargs["children"] = [msg]  # cannot be overwritten

        # create the alert
        alert = v.Alert(**kwargs)

        # add the alert to the app
        self.content.children = [alert] + self.content.children.copy()

        return self



from sepal_ui.message import ms

__all__ = ["MATRIX_NAMES", "TABLE_NAMES", "NO_VALUE", "SCHEMA"]

MATRIX_NAMES = ["src", "dst"]
TABLE_NAMES = ["code", "desc", "color"]

# Set a value for missing reclassifications
NO_VALUE = 999

SCHEMA = {
    "id": [ms.rec.table.schema.id, "number"],
    "code": [ms.rec.table.schema.code, "number"],
    "desc": [ms.rec.table.schema.description, "string"],
    "color": [ms.rec.table.schema.color, "hexa"],
}



import time

import ee

from sepal_ui.message import ms
from sepal_ui.scripts import utils as su


@su.need_ee
def wait_for_completion(task_descripsion, widget_alert=None):
    """
    Wait until the selected process is finished. Display some output information

    Args:
        task_descripsion (str): name of the running task
        widget_alert (v.Alert, optional): alert to display the output messages

    Return:
        (str): the final state of the task
    """
    state = "UNSUBMITTED"
    while state != "COMPLETED":

        # print in a widget
        if widget_alert:
            widget_alert.add_live_msg(ms.status.format(state))

        # wait 5 seconds
        time.sleep(5)

        # search for the task in task_list
        current_task = is_task(task_descripsion)
        state = current_task.state

        if state == "FAILED":
            raise Exception(ms.status.format(state))

    # print in a widget
    if widget_alert:
        widget_alert.add_live_msg(ms.status.format(state), "success")

    return state


@su.need_ee
def is_task(task_descripsion):
    """
    Search for the described task in the user Task list return None if nothing is found

    Args:
        task_descripsion (str): the task description

    Return:
        (ee.Task) : return the found task else None
    """

    current_task = None
    for task in ee.batch.Task.list():
        if task.config["description"] == task_descripsion:
            current_task = task
            break

    return current_task


@su.need_ee
def is_running(task_descripsion):
    """
    Search for the described task in the user Task list return None if nothing is currently running

    Args:
        task_descripsion (str): the task description

    Return:
        (ee.Task) : return the found task else None
    """

    current_task = is_task(task_descripsion)
    if current_task:
        if current_task.state not in ["RUNNING", "READY"]:
            current_task = None

    return current_task


@su.need_ee
def get_assets(folder=None, asset_list=[]):
    """
    Get all the assets from the parameter folder. every nested asset will be displayed.

    Args:
        folder (str): the initial GEE folder
        asset_list ([assets]| optional): extra element that you would like to add to the asset list

    Return:
        ([asset]): the asset list. each asset is a dict with 3 keys: 'type', 'name' and 'id'
    """
    # set the folder
    folder = folder if folder else ee.data.getAssetRoots()[0]["id"]

    # loop in the assets
    for asset in ee.data.listAssets({"parent": folder})["assets"]:
        if asset["type"] == "FOLDER":
            asset_list += [asset]
            asset_list = get_assets(asset["name"], asset_list)
        else:
            asset_list += [asset]

    return asset_list


@su.need_ee
def is_asset(asset_name, folder=None):
    """
    Check if the asset already exist in the user asset folder

    Args:
        asset_descripsion (str) : the descripsion of the asset
        folder (str): the folder of the glad assets

    Return:
        (bool): true if already in folder
    """

    # get the folder
    folder = folder or ee.data.getAssetRoots()[0]["id"]

    # get all the assets
    asset_list = get_assets(folder)

    # search for asset existance
    exist = False
    for asset in asset_list:
        if asset_name == asset["name"]:
            exist = True
            break

    return exist



# overwrite all the ipyvuetify widgets
from sepal_ui.sepalwidgets.sepalwidget import SepalWidget, TYPES
import ipyvuetify as v


# overwrite html
class Html(v.Html, SepalWidget):
    pass


# overwrite classes
_c_list = [
    c for c in dir(v.generated) if not c.startswith("__") and c != "VuetifyWidget"
]

for c in _c_list:

    class _tmp(getattr(v, c), SepalWidget):
        pass

    locals()[c] = _tmp
del _tmp

# import and/or overwrite with our customized widgets
from sepal_ui.sepalwidgets.widget import *
from sepal_ui.sepalwidgets.alert import *
from sepal_ui.sepalwidgets.btn import *
from sepal_ui.sepalwidgets.app import *
from sepal_ui.sepalwidgets.tile import *
from sepal_ui.sepalwidgets.inputs import *



import os
from pathlib import Path
from urllib.parse import urlparse
import string
import random
import math
import re
import warnings
from unidecode import unidecode
from functools import wraps
from itertools import product

import ee
from cryptography.fernet import Fernet
from matplotlib import colors as c
from deprecated.sphinx import deprecated

import sepal_ui

from .warning import SepalWarning


def hide_component(widget):
    """
    hide a vuetify based component

    Args:
        widget (v.VuetifyWidget): the widget to hide
    """

    if isinstance(widget, sepal_ui.sepalwidgets.sepalwidget.SepalWidget):
        widget.hide()

    elif "d-none" not in str(widget.class_):
        widget.class_ = str(widget.class_).strip() + " d-none"

    return widget


def show_component(widget):
    """
    show a vuetify based component

    Args:
        widget (v.VuetifyWidget): the widget to hide
    """

    if isinstance(widget, sepal_ui.sepalwidgets.sepalwidget.SepalWidget):
        widget.show()

    elif "d-none" in str(widget.class_):
        widget.class_ = widget.class_.replace("d-none", "")

    return widget


def create_download_link(pathname):
    """
    Create a clickable link to download the pathname target

    Args:
        pathname (str | pathlib.Path): the pathname th download

    Return:
        (str): the download link
    """

    # return the link if it's an absolute url
    if isinstance(pathname, str) and bool(urlparse(str(pathname)).netloc):
        return pathname

    # create a downloadable link from the jupyter node
    pathname = Path(pathname)
    try:
        download_path = pathname.relative_to(Path.home())
    except ValueError:
        download_path = pathname

    # I want to use the ipyurl lib to guess the url of the Jupyter server on the fly
    # but I don't really understand how it works
    # so here is an ugly fix only compatible with SEPAL
    link = f"https://sepal.io/api/sandbox/jupyter/files/{download_path}"

    return link


@deprecated(
    version="2.5.4",
    reason="This function makes no sense outside of create_download_link. It will be removed in the next minor version",
)
def is_absolute(url):
    """
    Check if the given URL is an absolute or relative path

    Args:
        url (str): the URL to test

    Return:
        (bool): True if absolute else False
    """
    return bool(urlparse(str(url)).netloc)


def random_string(string_length=3):
    """
    Generates a random string of fixed length.

    Args:
        string_length (int, optional): Fixed length. Defaults to 3.

    Return:
        (str): A random string
    """

    # random.seed(1001)
    letters = string.ascii_lowercase

    return "".join(random.choice(letters) for i in range(string_length))


def get_file_size(filename):
    """
    Get the file size as string of 2 digit in the adapted scale (B, KB, MB....)

    Args:
        filename (str | pathlib.Path): the path to the file to mesure

    Return:
        (str): the file size in a readable humanly readable
    """

    file_size = Path(filename).stat().st_size

    if file_size == 0:
        return "0B"

    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")

    i = int(math.floor(math.log(file_size, 1024)))
    s = file_size / (1024 ** i)

    return "{:.1f} {}".format(s, size_name[i])


def init_ee():
    """
    Initialize earth engine according to the environment.
    It will use the creddential file if the EE_PRIVATE_KEY env variable exist.
    Otherwise it use the simple Initialize command (asking the user to register if necessary)
    """

    # only do the initialization if the credential are missing
    if not ee.data._credentials:

        # if the decrypt key is available use the decript key
        if "EE_DECRYPT_KEY" in os.environ:

            # read the key as byte
            key = os.environ["EE_DECRYPT_KEY"].encode()

            # create the fernet object
            fernet = Fernet(key)

            # decrypt the key
            json_encrypted = Path(__file__).parent / "encrypted_key.json"
            with json_encrypted.open("rb") as f:
                json_decripted = fernet.decrypt(f.read()).decode()

            # write it to a file
            with open("ee_private_key.json", "w") as f:
                f.write(json_decripted)

            # connection to the service account
            service_account = "test-sepal-ui@sepal-ui.iam.gserviceaccount.com"
            credentials = ee.ServiceAccountCredentials(
                service_account, "ee_private_key.json"
            )
            ee.Initialize(credentials)

        # if in local env use the local user credential
        else:
            ee.Initialize()

    return


def catch_errors(alert, debug=False):
    """
    Decorator to execute try/except sentence
    and catch errors in the alert message.
    If debug is True then the error is raised anyway

    Params:
        alert (sw.Alert): Alert to display errors
        debug (bool): Wether to raise the error or not, default to false
    """

    def decorator_alert_error(func):
        @wraps(func)
        def wrapper_alert_error(*args, **kwargs):
            value = None
            try:
                value = func(*args, **kwargs)
            except Exception as e:
                alert.add_msg(f"{e}", type_="error")
                if debug:
                    raise e
            return value

        return wrapper_alert_error

    return decorator_alert_error


def need_ee(func):
    """
    Decorator to execute check if the object require EE binding.
    Trigger an exception if the connection is not possible.

    Params:
        func (obj): the object on which the decorator is applied
    """

    @wraps(func)
    def wrapper_ee(*args, **kwargs):

        # try to connect to ee
        try:
            init_ee()
        except Exception:
            raise Exception("This function needs an Earth Engine authentication")

        return func(*args, **kwargs)

    return wrapper_ee


def loading_button(alert=None, button=None, debug=False):
    """
    Decorator to execute try/except sentence and toggle loading button object.
    Designed to work within the Tile object, or any object that have a self.btn and self.alert set.

    Params:
        button (sw.Btn, optional): Toggled button
        alert (sw.Alert, optional): the alert to display the error message
        debug (bool, optional): wether or not the exception should stop the execution. default to False
    """

    def decorator_loading(func):
        @wraps(func)
        def wrapper_loading(self, *args, **kwargs):

            # set btn and alert
            # Change name of variable to assign it again in this scope
            button_ = self.btn if not button else button
            alert_ = self.alert if not alert else alert

            # Clean previous loaded messages in alert
            alert_.reset()

            button_.toggle_loading()  # Start loading
            value = None
            try:
                # Catch warnings in the process function
                with warnings.catch_warnings(record=True) as w_list:
                    value = func(self, *args, **kwargs)

                # Check if there are warnings in the function and append them
                # Use append msg as several warnings could be triggered
                if w_list:

                    # split the warning list
                    w_list_sepal = [
                        w for w in w_list if isinstance(w.message, SepalWarning)
                    ]

                    # display the sepal one
                    ms_list = [
                        f"{w.category.__name__}: {w.message.args[0]}"
                        for w in w_list_sepal
                    ]
                    [alert_.append_msg(ms, type_="warning") for ms in ms_list]

                    # only display them in the console if debug mode
                    if debug:

                        def custom_showwarning(w):
                            return warnings.showwarning(
                                message=w.message,
                                category=w.category,
                                filename=w.filename,
                                lineno=w.lineno,
                                line=w.line,
                            )

                        [custom_showwarning(w) for w in w_list]

            except Exception as e:
                alert_.add_msg(f"{e}", "error")
                if debug:
                    button_.toggle_loading()  # Stop loading button if there is an error
                    raise e

            button_.toggle_loading()  # Stop loading button

            return value

        return wrapper_loading

    return decorator_loading


def normalize_str(msg, folder=True):
    """
    Normalize an str to make it compatible with file naming (no spaces, special chars ...etc)

    Params:
        msg (str): the string to sanitise
        folder (optional|bool): if the name will be used for folder naming or for display. if display, <'> and < > characters will be kept

    Return:
        (str): the modified str
    """

    regex = "[^a-zA-Z\d\-\_]" if folder else "[^a-zA-Z\d\-\_\ ']"

    return re.sub(regex, "_", unidecode(msg))


def to_colors(in_color, out_type="hex"):
    """
    Transform any color type into a color in the specified output format
    avalable format: hex

    Args:
        in_color (str or tuple): It can be a string (e.g., 'red', '#ffff00', 'ffff00') or RGB tuple (e.g., (255, 127, 0)).
        out_type (str, optional): the type of the output color from ['hex']. default to 'hex'

    Returns:
        (str|tuple): The color in the specified format. default to black.
    """

    # list of the color function used for the translatio
    c_func = {"hex": c.to_hex}
    transform = c_func[out_type]

    out_color = "#000000"  # default black color

    if isinstance(in_color, tuple) and len(in_color) == 3:

        # rescale color if necessary
        if all(isinstance(item, int) for item in in_color):
            in_color = [c / 255.0 for c in in_color]

        return transform(in_color)

    else:

        # try to guess the color system
        try:
            return transform(in_color)
        except Exception:
            pass

        # try again by adding an extra # (GEE handle hex codes without #)
        try:
            return transform(f"#{in_color}")
        except Exception:
            pass

    return transform(out_color)


def switch(*params, debug=True, on_widgets=[], targets=[]):
    """
    Decorator to switch the state of input boolean parameters on class widgets or the
    class itself. If on_widgets is defined, it will switch the state of every widget
    parameter, otherwise it will change the state of the class (self). You can also set
    two decorators on the same function, one could affect the class and other the widgets.

    Args:
        *params (str): any boolean parameter of a SepalWidget.
        debug (bool): Whether trigger or not an Exception if the decorated function fails.
        on_widgets (list(widget_names,)|optional): List of widget names into the class
        targets (list(bool,)|optional); list of the target value (value taht will be set on switch. default to the inverse of the current state.

    """

    def decorator_switch(func):
        @wraps(func)
        def wrapper_switch(self, *args, **kwargs):

            widgets_len = len(on_widgets)
            targets_len = len(targets)

            # sanity check on targets and on_widgets
            if widgets_len and targets_len:
                if widgets_len != targets_len:
                    raise IndexError(
                        f'the length of "on_widgets" ({widgets_len}) is different from the length of "targets" ({targets_len})'
                    )

            # create the list of target values based on the target list
            # or the initial values of the widgets params
            # The first one is taken as reference
            if not targets_len:
                w = getattr(self, on_widgets[0]) if widgets_len else self
                targets_ = [bool(getattr(w, p)) for p in params]
            else:
                targets_ = targets

            if widgets_len:

                # Verify that the input elements are strings
                wrong_types = [
                    (w, type(w)) for w in on_widgets if not isinstance(w, str)
                ]

                if len(wrong_types):
                    errors = [
                        f"Received:{w_type} for widget: {w}."
                        for w, w_type in wrong_types
                    ]

                    raise TypeError(
                        f"All on_widgets list elements has to be strings. [{' '.join(errors)}]"
                    )

                missing_widgets = [w for w in on_widgets if not hasattr(self, w)]

                if missing_widgets:
                    raise Exception(
                        f"The provided {missing_widgets} widget(s) does not exist in the current class"
                    )

                def w_assign(bool_targets):

                    params_targets = [
                        (p, bool_targets[i]) for i, p in enumerate(params)
                    ]

                    for (w_name, p_t) in product(on_widgets, params_targets):
                        param, target = p_t
                        widget = getattr(self, w_name)
                        setattr(widget, param, target)

            else:

                def w_assign(bool_targets):
                    for i, p in enumerate(params):
                        setattr(self, p, bool_targets[i])

            # assgn the parameters to the target inverse
            w_assign([not t for t in targets_])

            # execute the function and catch errors
            try:
                func(self, *args, **kwargs)

            except Exception as e:
                if debug:
                    w_assign(targets_)
                    raise e

            # reassign the parameters to the targets
            w_assign(targets_)

        return wrapper_switch

    return decorator_switch


def next_string(string):
    """Create a string followed by an underscore and a consecutive number"""

    # if the string is already numbered the last digit is separeted from the rest of the string by an "_"
    split = string.split("_")
    end = split[-1]

    if end.isdigit():
        string = "_".join(split[:-1]) + f"_{int(end)+1}"
    else:
        string += "_1"

    return string



import json
from pathlib import Path
from traitlets import Any
from urllib.request import urlretrieve
import tempfile

import pandas as pd
import geopandas as gpd
from ipyleaflet import GeoJSON
import geemap
import ee

from sepal_ui.frontend.styles import AOI_STYLE
from sepal_ui.scripts import utils as su
from sepal_ui.scripts import gee
from sepal_ui.message import ms
from sepal_ui.model import Model

__all__ = ["AoiModel"]


class AoiModel(Model):
    """
    an Model object dedicated to the sorage and the manipulation of aoi.
    It is meant to be used with the AoiView object (embeded in the AoiTile).
    By using this you will be able to provide your application with aoi as an ee_object or a gdf, depending if you activated the ee binding or not.
    The class also provide insight on your aoi geometry.

    Args:
        gee (bool, optional): wether or not the aoi selector should be using the EarthEngine binding
        vector (str|pathlib.Path, optional): the path to the default vector object
        admin (int, optional): the administrative code of the default selection. Need to be GADM if ee==False and GAUL 2015 if ee==True.
        asset (str, optional): the default asset. Can only work if ee==True
        folder(str, optional): the init GEE asset folder where the asset selector should start looking (debugging purpose)

    .. deprecated:: 2.3.2
        'asset_name' will be used as variable to store 'ASSET' method info. To get the destination saved asset id, please use 'dst_asset_id' variable.

    """

    # ###########################################################################
    # ###                      dataset const                                  ###
    # ###########################################################################

    FILE = [
        Path(__file__).parents[1] / "scripts" / "gadm_database.csv",
        Path(__file__).parents[1] / "scripts" / "gaul_database.csv",
    ]
    "list(str): path to the GADM(0) and GAUL(1) database"

    CODE = ["GID_{}", "ADM{}_CODE"]
    "list(str): GADM(0) and GAUL(1) administrative codes key format"

    NAME = ["NAME_{}", "ADM{}_NAME"]
    "list(str): GADM(0) and GAUL(1) naming key format"

    ISO = ["GID_0", "ISO 3166-1 alpha-3"]
    "list(str): GADM(0) and GAUL(1) iso codes key"

    GADM_BASE_URL = "https://biogeo.ucdavis.edu/data/gadm3.6/gpkg/gadm36_{}_gpkg.zip"
    "str: the base url to download gadm maps"

    GADM_ZIP_DIR = Path.home() / "tmp" / "GADM_zip"
    GADM_ZIP_DIR.mkdir(parents=True, exist_ok=True)
    "pathlib.Path: the zip dir where we download the zips"

    GAUL_ASSET = "FAO/GAUL/2015/level{}"
    "str: the GAUL asset name"

    ASSET_SUFFIX = "aoi_"
    "str: the suffix to identify the asset in GEE"

    # ###########################################################################
    # ###                             const methods                           ###
    # ###########################################################################

    CUSTOM = ms.aoi_sel.custom
    "str: the word displayed for custom method in the relevant lang"

    ADMIN = ms.aoi_sel.administrative
    "str: the word displayed for admin method in the relevant lang"

    METHODS = {
        "ADMIN0": {"name": ms.aoi_sel.adm[0], "type": ADMIN},
        "ADMIN1": {"name": ms.aoi_sel.adm[1], "type": ADMIN},
        "ADMIN2": {"name": ms.aoi_sel.adm[2], "type": ADMIN},
        "SHAPE": {"name": ms.aoi_sel.vector, "type": CUSTOM},
        "DRAW": {"name": ms.aoi_sel.draw, "type": CUSTOM},
        "POINTS": {"name": ms.aoi_sel.points, "type": CUSTOM},
        "ASSET": {"name": ms.aoi_sel.asset, "type": CUSTOM},
    }
    "dict(str): the word displayed for all selection methods in the relevant lang"

    # ###########################################################################
    # ###                      widget related traitlets                       ###
    # ###########################################################################

    method = Any(None).tag(sync=True)
    "str: the currently selected method"

    point_json = Any(None).tag(sync=True)
    "dict: information that will be use to transform the csv into a gdf"

    vector_json = Any(None).tag(sync=True)
    "dict: information that will be use to transform the vector file into a gdf"

    geo_json = Any(None).tag(sync=True)
    "dict: the drawn geojson shape"

    admin = Any(None).tag(sync=True)
    "int: the admin number selected"

    asset_name = Any(None).tag(sync=True)
    "str: the asset name (only for GEE model)"

    name = Any(None).tag(sync=True)
    "str: the name of the file to create (used only in drawn shaped)"

    # ###########################################################################
    # ###                           model parameters                          ###
    # ###########################################################################

    ee = True
    "bool: either or not the model is bound to gee"

    folder = None
    "str: the folder name used in GEE related component, mainly used for debugging"

    alert = None
    "sepal_ui.sepalwidgets.Alert: the alert to display outputs"

    default_vector = None
    "(str|pathlib.path: the default vector file that will be used to produce the gdf. need to be readable by fiona and/or GDAL/OGR"

    default_asset = None
    "str: the default administrative area in GADM or GAUL norm"

    default_admin = None
    "str: the default asset name, need to point to a readable FeatureCollection"

    # ###########################################################################
    # ###                           model outputs                             ###
    # ###########################################################################

    dst_asset_id = None
    "str: the exported asset id"

    selected_feature = None
    "ee.Feature|GoeSeries: the Feature associated with a query"

    gdf = None
    "geopandas.GeoDataFrame: the geodataframe corresponding to the selected AOI"

    feature_collection = None
    "str: the name of the asset used to described the AOI"

    ipygeojson = None
    "ipyleaflet.GeoJSON: the representation of the AOI as a ipyleaflet layer"

    def __init__(
        self, alert, gee=True, vector=None, admin=None, asset=None, folder=None
    ):

        super().__init__()

        # the ee retated informations
        self.ee = gee
        if gee:
            su.init_ee()
            self.folder = folder or ee.data.getAssetRoots()[0]["id"]

        # the Alert used to display information
        self.alert = alert

        # set default values
        self.set_default(vector, admin, asset)

    def set_default(self, vector=None, admin=None, asset=None):
        """
        Set the default value of the object and create a gdf/feature_collection out of it

        Params:
            vector (str|pathlib.path, optional): the default vector file that will be used to produce the gdf. need to be readable by fiona and/or GDAL/OGR
            admin (str, optional): the default administrative area in GADM or GAUL norm
            asset (str, optional): the default asset name, need to point to a readable FeatureCollection

        Return:
            self
        """

        # save the default values
        self.default_vector = vector
        self.default_asset = self.asset_name = (
            {"pathname": asset, "column": "ALL", "value": None} if asset else None
        )
        self.default_admin = self.admin = admin

        # cast the vector to json
        self.vector_json = (
            {"pathname": str(vector), "column": "ALL", "value": None}
            if vector
            else None
        )

        # set the default gdf if possible
        if self.vector_json is not None:
            self.set_object("SHAPE")
        elif self.admin is not None:
            self.set_object("ADMIN0")  # any level will work
        elif self.asset_name is not None:
            self.set_object("ASSET")

        return self

    def set_object(self, method=None):
        """
        set the object (gdf/featurecollection) based on the model inputs. The method can be manually overwrite

        Args:
            method (str, optional): a model loading method

        Return:
            self
        """

        # clear the model output if existing
        self.clear_output()

        # overwrite self.method
        self.method = method or self.method

        if self.method in ["ADMIN0", "ADMIN1", "ADMIN2"]:
            self._from_admin(self.admin)
        elif self.method == "POINTS":
            self._from_points(self.point_json)
        elif self.method == "SHAPE":
            self._from_vector(self.vector_json)
        elif self.method == "DRAW":
            self._from_geo_json(self.geo_json)
        elif self.method == "ASSET":
            self._from_asset(self.asset_name)
        else:
            raise Exception(ms.aoi_sel.no_inputs)

        self.alert.add_msg(ms.aoi_sel.complete, "success")

        return self

    def _from_asset(self, asset_name):
        """set the ee.FeatureCollection output from an existing asset"""

        if not (asset_name["pathname"]):
            raise Exception("Please select an asset.")

        if asset_name["column"] != "ALL":
            if asset_name["value"] is None:
                raise Exception("Please select a value.")

        self.name = Path(asset_name["pathname"]).stem.replace(self.ASSET_SUFFIX, "")
        ee_col = ee.FeatureCollection(asset_name["pathname"])

        if asset_name["column"] != "ALL":

            column = asset_name["column"]
            value = asset_name["value"]
            ee_col = ee_col.filterMetadata(column, "equals", value)
            self.name = f"{self.name}_{column}_{value}"

        # set the feature collection
        self.feature_collection = ee_col

        # create a gdf form te feature_collection
        # cannot be used before geemap 0.8.17 (not released)
        # self.gdf = geemap.ee_to_geopandas(self.feature_collection)
        self.gdf = gpd.GeoDataFrame.from_features(
            self.feature_collection.getInfo()["features"]
        ).set_crs(epsg=4326)

        # set the name

        return self

    def _from_points(self, point_json):
        """set the object output from a csv json"""

        if not all(point_json.values()):
            raise Exception("All fields are required, please fill them.")

        # cast the pathname to pathlib Path
        point_file = Path(point_json["pathname"])

        # check that the columns are well set
        values = [v for v in point_json.values()]
        if not len(values) == len(set(values)):
            raise Exception(ms.aoi_sel.duplicate_key)

        # create the gdf
        df = pd.read_csv(point_file, sep=None, engine="python")
        self.gdf = gpd.GeoDataFrame(
            df,
            crs="EPSG:4326",
            geometry=gpd.points_from_xy(
                df[point_json["lng_column"]], df[point_json["lat_column"]]
            ),
        )

        # set the name
        self.name = point_file.stem

        if self.ee:
            # transform the gdf to ee.FeatureCollection
            self.feature_collection = geemap.geojson_to_ee(self.gdf.__geo_interface__)

            # export as a GEE asset
            self.export_to_asset()

        return self

    def _from_vector(self, vector_json):
        """set the object output from a vector json"""

        if not (vector_json["pathname"]):
            raise Exception("Please select a file.")

        if vector_json["column"] != "ALL":
            if vector_json["value"] is None:
                raise Exception("Please select a value.")

        # cast the pathname to pathlib Path
        vector_file = Path(vector_json["pathname"])

        # create the gdf
        self.gdf = gpd.read_file(vector_file).to_crs("EPSG:4326")

        # set the name using the file stem
        self.name = vector_file.stem

        # filter it if necessary
        if vector_json["value"] is not None:
            self.gdf = self.gdf[self.gdf[vector_json["column"]] == vector_json["value"]]
            self.name = f"{self.name}_{vector_json['column']}_{vector_json['value']}"

        if self.ee:
            # transform the gdf to ee.FeatureCollection
            self.feature_collection = geemap.geojson_to_ee(self.gdf.__geo_interface__)

            # export as a GEE asset
            self.export_to_asset()

        return self

    def _from_geo_json(self, geo_json):
        """set the gdf output from a geo_json"""

        if not geo_json:
            raise Exception("Please draw a shape in the map")

        # remove the style property from geojson as it's not recognize by geopandas and gee
        for feat in geo_json["features"]:
            if "style" in feat["properties"]:
                del feat["properties"]["style"]

        # create the gdf
        self.gdf = gpd.GeoDataFrame.from_features(geo_json).set_crs(epsg=4326)

        # normalize the name
        self.name = su.normalize_str(self.name)

        if self.ee:
            # transform the gdf to ee.FeatureCollection
            self.feature_collection = geemap.geojson_to_ee(self.gdf.__geo_interface__)

            # export as a GEE asset
            self.export_to_asset()
        else:
            # save the geojson in downloads
            path = Path("~", "downloads", "aoi").expanduser()
            path.mkdir(
                exist_ok=True, parents=True
            )  # if nothing have been run the downloads folder doesn't exist
            self.gdf.to_file(path / f"{self.name}.geojson", driver="GeoJSON")

        return self

    def _from_admin(self, admin):
        """Set the object according to given an administrative number in the GADM norm. The object will be projected in EPSG:4326"""

        if not admin:
            raise Exception("Select an administrative layer")

        # get the admin level corresponding to the given admin code
        df = pd.read_csv(self.FILE[self.ee])

        # extract the first element that include this administrative code and set the level accordingly
        is_in = df.filter([self.CODE[self.ee].format(i) for i in range(3)]).isin(
            [admin]
        )

        if not is_in.any().any():
            raise Exception("The code is not in the database")
        else:
            index = 3 if self.ee else -1
            level = (
                is_in[~((~is_in).all(axis=1))].idxmax(1).iloc[0][index]
            )  # the character that contains the index

        if self.ee:

            # get the feature_collection
            self.feature_collection = ee.FeatureCollection(
                self.GAUL_ASSET.format(level)
            ).filter(ee.Filter.eq(f"ADM{level}_CODE", admin))

            # transform it into gdf
            # cannot be used before geemap 0.8.17 (not released)
            # self.gdf = geemap.ee_to_geopandas(self.feature_collection)
            self.gdf = gpd.GeoDataFrame.from_features(
                self.feature_collection.getInfo()["features"]
            ).set_crs(epsg=4326)

        else:
            # save the country iso_code
            iso_3 = admin[:3]

            # download the geopackage in a tmp directory
            with tempfile.TemporaryDirectory() as tmp_dir:

                zip_file = Path(tmp_dir) / f"{iso_3}.zip"

                # get the zip from GADM server only the ISO_3 code need to be used
                urlretrieve(self.GADM_BASE_URL.format(iso_3), zip_file)

                # read the geopackage
                layer_name = f"gadm36_{iso_3}_{level}"
                level_gdf = gpd.read_file(
                    f"{zip_file}!gadm36_{iso_3}.gpkg", layer=layer_name
                )

            # get the exact admin from this layer
            self.gdf = level_gdf[level_gdf[self.CODE[self.ee].format(level)] == admin]

        # set the name using the layer
        r = df[df[self.CODE[self.ee].format(level)] == admin].iloc[0]
        names = [
            su.normalize_str(r[self.NAME[self.ee].format(i)])
            if i
            else r[self.ISO[self.ee]]
            for i in range(int(level) + 1)
        ]
        self.name = "_".join(names)

        return self

    def clear_output(self):
        """
        Clear the output of the aoi selector without changing the traits and/or the parameters.

        Return:
            self
        """

        # reset the outputs
        self.gdf = None
        self.feature_collection = None
        self.ipygeojson = None
        self.selected_feature = None
        self.dst_asset_id = None

        return self

    def clear_attributes(self):
        """
        Return all attributes to their default state.
        Set the default setting as current object.

        Return:
            self
        """

        # keep the default
        admin = self.default_admin
        vector = self.default_vector
        asset = self.default_asset

        # delete all the traits
        [setattr(self, attr, None) for attr in self.trait_names()]

        # reset the outputs
        self.clear_output()

        # reset the default
        self.set_default(vector, admin, asset)

        return self

    def get_columns(self):
        """
        Retrieve the columns or variables from self excluding geometries and gee index.

        Return:
            ([str]): sorted list of column names
        """

        if self.gdf is None:
            raise Exception("You must set the gdf before interacting with it")

        if self.ee:
            aoi_ee = ee.Feature(self.feature_collection.first())
            columns = aoi_ee.propertyNames().getInfo()
            list_ = [
                col for col in columns if col not in ["system:index", "Shape_Area"]
            ]
        else:
            list_ = list(set(["geometry"]) ^ set(self.gdf.columns.to_list()))

        return sorted(list_)

    def get_fields(self, column):
        """
        Retrieve the fields from a column

        Args:
            column (str): A column name to query over the asset

        Return:
            ([str]): sorted list of fields value

        """

        if self.gdf is None:
            raise Exception("You must set the gdf before interacting with it")

        if self.ee:
            fields = self.feature_collection.distinct(column).aggregate_array(column)
            list_ = fields.getInfo()
        else:
            list_ = self.gdf[column].to_list()

        return sorted(list_)

    def get_selected(self, column, field):
        """
        Select an ee object based on selected column and field.

        Return:
            (ee.Feature|GoeSeries): the Feature associated with the query
        """

        if self.gdf is None:
            raise Exception("You must set the gdf before interacting with it")

        if self.ee:
            selected_feature = self.feature_collection.filterMetadata(
                column, "equals", field
            )
        else:
            selected_feature = self.gdf[self.gdf[column] == field]

        return selected_feature

    def total_bounds(self):
        """
        Reproduce the behaviour of the total_bounds method from geopandas

        Return:
            (tuple): minxx, miny, maxx, maxy
        """

        if self.ee:
            ee_bounds = self.feature_collection.geometry().bounds().coordinates()
            coords = ee_bounds.get(0).getInfo()
            ll, ur = coords[0], coords[2]
            bounds = ll[0], ll[1], ur[0], ur[1]
        else:
            bounds = self.gdf.total_bounds

        return bounds

    def export_to_asset(self):
        """
        Export the feature_collection as an asset (only for ee model)

        Return:
            self
        """

        asset_name = self.ASSET_SUFFIX + self.name
        asset_id = str(Path(self.folder, asset_name))

        self.dst_asset_id = asset_id

        # check if the table already exist
        if asset_id in [a["name"] for a in gee.get_assets(self.folder)]:
            return self

        # check if the task is running
        if gee.is_running(asset_name):
            return self

        # run the task
        task_config = {
            "collection": self.feature_collection,
            "description": asset_name,
            "assetId": asset_id,
        }

        task = ee.batch.Export.table.toAsset(**task_config)
        task.start()

        return self

    def get_ipygeojson(self):
        """
        Converts current geopandas object into ipyleaflet GeoJSON

        Return:
            (GeoJSON): the geojson layer of the aoi gdf
        """

        if self.gdf is None:
            raise Exception("You must set the gdf before converting it into GeoJSON")

        # read the data from geojson and add the name as a property of the shape
        # useful when handler are added from ipyleaflet
        data = json.loads(self.gdf.to_json())
        for f in data["features"]:
            f["properties"]["name"] = self.name

        # create a GeoJSON object
        self.ipygeojson = GeoJSON(
            data=data, style=AOI_STYLE, name="aoi", attribution="SEPA(c)"
        )

        return self.ipygeojson



from traitlets import HasTraits
import json

from ipywidgets import dlink

__all__ = ["Model"]


class Model(HasTraits):
    """
    Model object to store information of your computation (input, output.. etc)
    The Model structure is based on traitlets and embed function for export or import in json format.
    The model traintlets can be bind to any widget field.

    """

    def __repr__(self):
        """Method to represent the Model objects as a string"""

        args = ", ".join([f"{k}={v}" for k, v in self.export_data().items()]).strip()

        return f"{self.__class__.__name__}({args})"

    def export_data(self):
        """
        Export a json description of all the object traitlets
        Note that the members will be ignored

        Return:
            (dict): the serialized traitlets
        """

        return self.__dict__["_trait_values"]

    def import_data(self, data):
        """
        Import the model (i.e. the traitlets) from a json file

        Params:
            json (dict): the json ditionary of the model

        Return:
            self
        """

        if type(data) == str:
            data = json.loads(data)

        for k, v in data.items():
            getattr(self, k)  # to raise an error if the json is malformed
            setattr(self, k, v)

    def bind(self, widget, trait):
        """
        Binding a widget input 'v_model' trait to a trait of the model.
        The binding will be unidirectionnal for the sake of some custom widget that does not support it.
        This wrapper avoid to import the ipywidgets lib everywhere and reduce the number of parameters
        Some existence check are also performed and will throw an error if the trait doesn't exist

        Params:
            widget (v.widget): any input widget with a v_model trait
            trait (str): the name of a trait of the current model

        Return:
            self
        """

        # check trait existence
        getattr(self, trait)

        # bind them
        dlink((widget, "v_model"), (self, trait))

        # maybe I would add the possiblity to add an alert to display stuff to the user with the same options as in alert.bind

        return self



import json
from types import SimpleNamespace
from pathlib import Path
from collections import abc
from deepdiff import DeepDiff


class Translator(SimpleNamespace):
    """
    The translator is a SimpleNamespace of Simplenamespace. It reads 2 Json files, the first one being the source language (usually English) and the second one the target language.
    It will replace in the source dictionary every key that exist in both json dictionaries. Following this procedure, every message that is not translated can still be accessed in the source language.
    To access the dictionary keys, instead of using [], you can simply use key name as in an object ex: translator.first_key.secondary_key.
    There are no depth limits, just respect the snake_case convention when naming your keys in the .json files.

    Args:
        json_folder (str | pathlib.Path): The folder where the dictionaries are stored
        target_lan (str): The language code of the target lang (it should be the same as the target dictionary)
        default_lan (str, optional): The language code of the source lang. default to "en"(it should be the same as the source dictionary)
        file_pattern (str, optional): The file basename of the dictionary without extention. default to "locale".
    """

    FORBIDDEN_KEYS = ["default_dict", "target_dict", "in", "class"]
    "list(str): list of the forbidden keys, using one of them in a translation dict will throw an error"

    target_dict = {}
    "(dict): the target language dictionary"

    default_dict = {}
    "dict: the source language dictionary"

    keys = None
    "all the keys can be acceced as attributes"

    def __init__(
        self, json_folder, target_lan, default_lan="en", file_pattern="locale"
    ):

        super().__init__()

        if type(json_folder) == str:
            json_folder = Path(json_folder)

        # reading the default dict
        source_path = json_folder / default_lan / f"{file_pattern}.json"
        self.default_dict = self.sanitize(json.loads(source_path.read_text()))

        # create a dictionary in the target language
        target_path = json_folder / target_lan / f"{file_pattern}.json"
        self.target_dict = self.default_dict.copy()

        if target_path.is_file():
            self.target_dict = self.sanitize(json.loads(target_path.read_text()))
        else:
            print(f'No json file is provided for "{target_lan}", fallback to "en"')

        # create the composite dictionary
        ms_dict = self._update(self.default_dict, self.target_dict)

        # check if forbidden keys are being used
        [self.search_key(ms_dict, k) for k in self.FORBIDDEN_KEYS]

        # transform it into a json str
        ms_json = json.dumps(ms_dict)

        # unpack the json as a simple namespace
        ms = json.loads(ms_json, object_hook=lambda d: SimpleNamespace(**d))

        for k, v in ms.__dict__.items():
            setattr(self, k, getattr(ms, k))

    @classmethod
    def search_key(cls, d, key):
        """
        Search a specific key in the d dictionary and raise an error if found

        Args:
            d (dict): the dictionary to study
            key (str): the key to look for
        """

        for k, v in d.items():
            if isinstance(v, abc.Mapping):
                cls.search_key(v, key)
            else:
                if k == key:
                    raise Exception(
                        f"You cannot use the key {key} in your translation dictionary"
                    )

        return

    @classmethod
    def sanitize(cls, d):
        """
        Identify numbered dictionnaries embeded in the dict and transform them into lists

        This function is an helper to prevent deprecation after the introduction of pontoon for translation.
        The user is now force to use keys even for numbered lists. SimpleNamespace doesn't support integer indexing so this function will transform back this "numbered" dictionnary (with integer keys) into lists.

        Args:
            d (dict): the dictionnary to sanitize

        Return:
            (dict): the sanitized dictionnary
        """

        ms = d.copy()

        # create generator based on input type
        if isinstance(ms, dict):
            gen = ms.items()
        elif isinstance(ms, list):
            gen = enumerate(ms)

        # loop into the keys of the dict modify them
        for k, v in gen:
            if isinstance(v, dict):
                tmp = v
                if all([k.isnumeric() for k in tmp]):
                    tmp = list(tmp.values())
                ms[k] = cls.sanitize(tmp)
            else:
                ms[k] = v

        return ms

    def _update(self, d, u):
        """
        Update the fallback dictionnaire (d) values with the keys that exist in the target (u) dictionnaire

        Args:
            d (dict): The fallback dictionary
            u (dict): the target dctionnary

        Return:
            ms (dict): The updated dictionnay
        """

        ms = d.copy()

        for k, v in u.items():
            if isinstance(v, abc.Mapping):
                ms[k] = self._update(d.get(k, {}), v)
            else:
                ms[k] = v

        return ms

    def missing_keys(self):
        """
        this function is intended for developer use only
        print the list of the missing keys in the target dictionnairie

        Return:
            (str): the list of missing keys
        """

        # find all the missing keys
        try:
            ddiff = DeepDiff(self.default_dict, self.target_dict)[
                "dictionary_item_removed"
            ]
        except Exception:
            ddiff = ["All messages are translated"]

        return "\n".join(ddiff)
